/*
    Copyright (C) 1999-2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: alsa_device.cc,v 1.65 2000/11/04 16:35:47 pbd Exp $
*/

#include <math.h>
#include <stdio.h>
#include <memory.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/poll.h>

#include <pbd/error.h>

#include "alsa_device.h"
#include "audiothing.h"
#include "hammerfall.h"
#include "generic_hw.h"
#include "configuration.h"
#include "utils.h"

// #define ALSA_MCD_WITH_GAIN


using namespace ALSA;

template<class SampleType> void
adjust_gain (SampleType *sample, size_t frames, size_t skip, float scale)

{
	while (frames--) {
		*sample = (SampleType) ((float) *sample * scale);
		sample += skip;
	}
}

void *
MultiChannelDevice::start_thread (void *arg)

{
	return ((MultiChannelDevice *) arg)->main ();
}

MultiChannelDevice::MultiChannelDevice (const string &name,
					size_t rate,
					size_t frames_per_cycle,
					bool real_time)

	: QMThread ("multichannel device", start_thread, this, real_time, 20),
	  audiothing_requests (16)

{
	int err;
	snd_pcm_params_info_t params_info;

	_ok = false;

	if (!thread_ok()) {
		return;
	}

	if ((err = snd_pcm_open (&playback_handle, (char *) name.c_str(), SND_PCM_STREAM_PLAYBACK, 0)) < 0) {
		error << "ALSA-MCD: Cannot open PCM device "
		      << name
		      << endmsg;
		return;
	}

	if ((err = snd_pcm_open (&capture_handle, (char *) name.c_str(), SND_PCM_STREAM_CAPTURE, 0)) < 0) {
		error << "ALSA-MCD: Cannot open PCM device "
		      << name
		      << endmsg;
		return;
	}

	audiothing = 0;
	ctl_handle = 0;
	overs[0] = 0;
	overs[1] = 0;
	over_length = 0;
	over_length_short = 0;
	over_length_long = 0;
	_peak_input_power = 0;
	_peak_output_power = 0;
	gain = 0;
	hw = 0;
	alsa_name = name;
	independent_go = false;
	input_monitor_mask = 0;
	last_mask = 0;
	input_meter_mask = 0;
	output_meter_mask = 0;
	all_monitor_in = false;
	fragment_bytes = 0;
	_nfragments = 0;
	_nchannels = 0;
	playback_addr = 0;
	capture_addr = 0;
	silent = 0;

	pthread_mutex_init (&restart_lock, NULL);
	pthread_cond_init (&_restart, NULL);

	/* XXX 0 may not be the right value */

	playback_frag = 0;
	capture_frag = 0;
	read_via_copy = 0;
	write_via_copy = 0;
	channel_copy = 0;
	true_copy = 0;

	if (check_card_type()) {
		return;
	}

	/* Look at the first capture subdevice, and ask for the number
	   of channels it supports.

	   XXX may need better logic here w.r.t the subdevices.
	*/

	memset (&params_info, 0, sizeof (params_info));

	if ((err = snd_pcm_params_info (capture_handle, &params_info)) < 0) {
		error << "ALSA::MCD: cannot get capture channel info ("
		      << snd_strerror (err)
		      << ')'
		      << endmsg;
		return;
	}
					 
	_nchannels = params_info.max_channels;

	/* Now figure out the best possible sample format we can use
	 */

	if (params_info.formats & SND_PCM_FMT_S32_LE) {
		_sample_format = SND_PCM_SFMT_S32_LE;
	} else if (params_info.formats & SND_PCM_FMT_S16_LE) {
		_sample_format = SND_PCM_SFMT_S16_LE;
	} else {
		error << "Sorry. The audio interface \""
		      << alsa_name
		      << "\" doesn't support either of the two sample formats that ardour can use."
		      << endmsg;
		return;
	}

	/* XXX what if these are different for playback ? */

	independent_go = !(params_info.flags & SND_PCM_INFO_SYNC_START);
	noninterleaved = (params_info.flags & SND_PCM_INFO_NONINTERLEAVED);

	/* set up the bit pattern that is used to record which
	   channels require action on every cycle. any bits that
	   are not set after the currently connected "AudioThing"
	   has done its stuff are either monitored or silenced.
	*/

	for (int n = _nchannels - 1; n >= 0; n--) {
		channel_done_bits |= (1<<n);
	}

	playback_addr = new char **[_nchannels];
	capture_addr = new char **[_nchannels];
	silent = new bool * [_nchannels];
	out_cnt = new size_t [_nchannels];
	in_cnt = new size_t [_nchannels];
	folding = new bool [_nchannels];

	read_via_copy = new ReadCopyFunction [_nchannels];
	write_via_copy = new WriteCopyFunction [_nchannels];
	channel_copy = new CopyCopyFunction [_nchannels];
	true_copy = new CopyCopyFunction [_nchannels];

	memset (playback_addr, 0, sizeof (char **) * _nchannels);
	memset (capture_addr, 0, sizeof (char **) * _nchannels);
	
	for (size_t chn = 0; chn < _nchannels; chn++) {
		channel_copy[chn] = memcpy_fake;
		true_copy[chn] = memcpy_fake;
		folding[chn] = false;
	}

	memset (out_cnt, 0, sizeof (size_t) * _nchannels);
	memset (in_cnt, 0, sizeof (size_t) * _nchannels);
	memset (read_via_copy, 0, sizeof (ReadCopyFunction) * _nchannels);
	memset (write_via_copy, 0, sizeof (WriteCopyFunction) * _nchannels);

	overs[0] = new size_t [_nchannels];
	overs[1] = new size_t [_nchannels];
	over_length = new size_t [_nchannels];
	_peak_input_power = new float [_nchannels];
	_peak_output_power = new float [_nchannels];
	gain = new float[_nchannels];

	_xfade_samples = 0;
	xfade_out = 0;
	xfade_in = 0;

	set_parameters (frames_per_cycle, rate);

	if (!independent_go) {
		snd_pcm_link (capture_handle, playback_handle);
	}

	reset_all_overs ();
	reset_all_peaks ();
	reset_gain ();

	hw_specific ();

	_ok = true;
}

MultiChannelDevice::~MultiChannelDevice ()

{
	snd_pcm_close (capture_handle);
	snd_pcm_close (playback_handle);

	if (_peak_input_power) {
		delete [] _peak_input_power;
	}

	if (_peak_output_power) {
		delete [] _peak_output_power;
	}

	if (playback_addr) {
		for (size_t i = 0; i < _nchannels; i++) {
			if (playback_addr[i]) {
				delete [] playback_addr[i];
			}
		}
		delete [] playback_addr;
	}

	if (capture_addr) {
		for (size_t i = 0; i < _nchannels; i++) {
			if (capture_addr[i]) {
				delete [] capture_addr[i];
			}
		}
		delete [] capture_addr;
	}

	if (silent) {
		for (size_t i = 0; i < _nchannels; i++) {
			delete [] silent[i];
		}
		delete [] silent;
	}

	if (overs[0]) {
		delete overs[0];
	}

	if (overs[1]) {
		delete overs[1];
	}

	if (over_length) {
		delete [] over_length;
	}

	if (xfade_out) {
		delete [] xfade_out;
	}

	if (xfade_in) {
		delete [] xfade_in;
	}

	if (folding) {
		delete [] folding;
	}

	if (out_cnt) {
		delete [] out_cnt;
	}

	if (in_cnt) {
		delete [] in_cnt;
	}

	if (read_via_copy) {
		delete [] read_via_copy;
	}

	if (write_via_copy) {
		delete [] write_via_copy;
	}

	if (channel_copy) {
		delete [] channel_copy;
	}

	if (true_copy) {
		delete [] true_copy;
	}
}

int 
MultiChannelDevice::set_parameters (size_t frames_per_cycle, size_t rate)

{
	switch (_sample_format) {
	case SND_PCM_SFMT_S32_LE:

		/* XXX must handle the n-bits of 24-in-32 problems here */

		if (Config->max_level) {
			max_level = Config->max_level;
		} else {
			max_level = INT_MAX;
		}

		if (Config->min_level) {
			min_level = Config->min_level;
		} else {
			min_level = INT_MIN;
		}
		break;

	case SND_PCM_SFMT_S16_LE:

		if (Config->max_level) {
			max_level = Config->max_level;
		} else {
			max_level = SHRT_MAX;
		}

		if (Config->min_level) {
			min_level = Config->min_level;
		} else {
			min_level = SHRT_MIN;
		}
		break;
	}

	_frame_rate = rate;
	_sample_bytes = snd_pcm_format_physical_width (_sample_format) / 8;
	_frames_per_cycle = frames_per_cycle;
	_bytes_per_cycle = _sample_bytes * _frames_per_cycle;

	if (configure (SND_PCM_STREAM_CAPTURE)) {
		error << "ALSA-MCD: cannot configure capture channel" 
		      << endmsg;
		return -1;
	}

	if (configure (SND_PCM_STREAM_PLAYBACK)) {
		error << "ALSA-MCD: cannot configure playback channel" 
		      << endmsg;
		return -1;
	}

	return 0;
}	

int
MultiChannelDevice::configure (int which_stream)

{
	snd_pcm_params_t params;
	snd_pcm_setup_t setup;
	snd_pcm_t *handle;
	int err;

	if (which_stream == SND_PCM_STREAM_PLAYBACK) {
		handle = playback_handle;
	} else {
		handle = capture_handle;
	}

	memset(&params, 0, sizeof(params));

	params.format.sfmt = _sample_format;
	params.format.rate = _frame_rate;
	params.format.channels = _nchannels; 
	params.ready_mode = SND_PCM_READY_FRAGMENT;
	params.start_mode = SND_PCM_START_EXPLICIT;
	params.xfer_mode = 
		noninterleaved ? 
		SND_PCM_XFER_NONINTERLEAVED : SND_PCM_XFER_INTERLEAVED;
	params.mmap_shape = 
		noninterleaved ? 
		SND_PCM_MMAP_NONINTERLEAVED : SND_PCM_MMAP_INTERLEAVED;
	params.xrun_mode = SND_PCM_XRUN_NONE;
	params.frag_size = _frames_per_cycle;
	params.avail_min = params.frag_size;  // poll(2) returns with 1 frag

	if (fragment_bytes == 0) {

		/* ask for 3 fragments, because thats optimal.
		   We might get 2, 3 or 4. It all depends.
		*/
		
		params.buffer_size = params.frag_size * 3;

	} else {
		/* use the value we got from the first channel for
		   the second.
		*/

		params.buffer_size = _frames_per_cycle * _nfragments;
	}

	if ((err = snd_pcm_params (handle, &params)) < 0) {
		error << "ALSA-MCD: stream "
		      << which_stream
		      << " on "
		      << alsa_name
		      << " parameter error: " 
		      << snd_strerror(err) 
		      << endmsg;
		return -1;
	}
	
	if (which_stream == SND_PCM_STREAM_CAPTURE) {
		if ((err = snd_pcm_mmap (capture_handle)) < 0) {
			error << "ALSA::Mapped_Device: cannot map capture device ("
			      << snd_strerror (err)
			      << ')'
			      << endmsg;
			return -1;
		} 

	} else {
		if ((err = snd_pcm_mmap (playback_handle)) < 0) {
			error << "ALSA::Mapped_Device: cannot map playback device ("
			      << snd_strerror (err)
			      << ')'
			      << endmsg;
			return -1;
		} 
	}

	if ((err = snd_pcm_setup (handle, &setup)) < 0) {
		error << "ALSA-MCD: cannot get channel setup info ("
		      << snd_strerror (err)
		      << ')'
		      << endmsg;
		return -1;
	}

	/* do this stuff just for the first stream that 
	   we configure, since fragment_bytes, _nfragments etc.
	   are assumed to be the same for both streams.
	*/

	if (fragment_bytes == 0) {

		/* The fragment size is non-negotiable.
		 */

		if (setup.frag_size != _frames_per_cycle) {
			error << "ALSA::MCD: requested fragment of "
			      << _frames_per_cycle
			      << " frames but got "
			      << setup.frag_size
			      << " frames"
			      << endmsg;
			return -1;
		}

		fragment_bytes = setup.frag_size * 
			(snd_pcm_format_physical_width 
			 (setup.format.sfmt) / 8) * _nchannels;

		_nfragments = setup.frags;
	}

	/* I cannot imagine a device with playback and capture
	   having different interleave status. So, yes, we do this
	   twice, and use the final value(s).
	*/

	if (setup.xfer_mode == SND_PCM_XFER_INTERLEAVED) {
		interleaved = true;
		interleave_unit = snd_pcm_format_physical_width (setup.format.sfmt) / 8;
		interleave_skip = interleave_unit * _nchannels;
	} else {
		interleaved = false;
		interleave_unit = 0;  /* NOT USED */
		interleave_skip = snd_pcm_format_physical_width (setup.format.sfmt) / 8;
	}

	if (which_stream == SND_PCM_STREAM_PLAYBACK) {
		for (size_t chn = 0; chn < _nchannels; chn++) {
			silent[chn] = new bool [_nfragments];
			memset (silent[chn], 1, sizeof (bool) * _nfragments);
		}
	}

	snd_pcm_channel_area_t areas[_nchannels];
	snd_pcm_channel_area_t stopped_areas[_nchannels];
	
	if ((err = snd_pcm_mmap_get_areas (handle, stopped_areas, areas)) < 0) {
		error << "ALSA-HW: mmap areas info error for stream "
		      << which_stream
		      << " on "
		      << alsa_name
		      << snd_strerror(err)
		      << endmsg;
		return -1;
	}

	char ****addrptr;  // !!!!!
	
	if (which_stream == SND_PCM_STREAM_CAPTURE) {
		addrptr = &capture_addr;
	} else {
		addrptr = &playback_addr;
	}

	for (size_t chn = 0; chn < _nchannels; chn++) {

		(*addrptr)[chn] = new char *[_nfragments];
		memset ((*addrptr)[chn], 0, sizeof (char *) * _nfragments);

		for (size_t frag = 0; frag < _nfragments; frag++) {
			(*addrptr)[chn][frag] = (((char *) areas[chn].addr) + (areas[chn].first / 8));
			
			if (interleaved) {
				(*addrptr)[chn][frag] += (frag * fragment_bytes);
			} else {
				(*addrptr)[chn][frag] += (frag * _channel_fragment_bytes);
			}
		}
	}

	return 0;
}

int
MultiChannelDevice::check_card_type ()

{
	int err;
	struct snd_ctl_hw_info ctlinfo;

	if ((err = snd_ctl_open (&ctl_handle, (char *) alsa_name.c_str())) < 0) {
		error << "control open ("
		      << alsa_name
		      << "): "
		      << snd_strerror(err)
		      << endmsg;
		return -1;
	}
	
	if ((err = snd_ctl_hw_info(ctl_handle, &ctlinfo)) < 0) {
		error << "control hardware info ("
		      << alsa_name
		      << "): "
		      << snd_strerror(err)
		      << endmsg;
		snd_ctl_close(ctl_handle);
		return -1;
	}

	card_type = ctlinfo.type;
	return check_capabilities ();
}

int
MultiChannelDevice::hw_specific ()

{
	int err;

	_hw_monitoring = false;

	switch (card_type) {
	case SND_CARD_TYPE_HAMMERFALL:
	case SND_CARD_TYPE_HAMMERFALL_LIGHT:
		if ((err = hammerfall_hardware ()) != 0) {
			return err;
		}
		break;

	default:
		if ((err = generic_hardware ()) != 0) {
			return err;
		}
		break;
	}

	if (has_hw_monitoring()) {
		_hw_monitoring = true;
	}

	return 0;
}

int
MultiChannelDevice::hammerfall_hardware ()

{
	hw = new Hammerfall (this, ctl_handle);
	return 0;
}

int
MultiChannelDevice::generic_hardware ()

{
	hw = new GenericHardware (this, ctl_handle);
	return 0;
}

int
MultiChannelDevice::io_loop ()

{
	int fd;
	ssize_t avail;
	bool xrun_detected;

	fd = snd_pcm_poll_descriptor (capture_handle);

	while (!work_no_more ()) {

		struct pollfd pfd;
		pfd.fd = fd;
		pfd.events = POLLIN | POLLERR;

	  again:
		if (poll (&pfd, 1, 1000) < 0) {
			if (errno == EINTR) {
				// this happens mostly when run
				// under gdb, or when exiting due to a signal
				goto again;
			}

			error << "ALSA::Device: poll call failed ("
			      << strerror (errno)
			      << ')'
			      << endmsg;
			return -1;
		}

		if (pfd.revents & POLLERR) {
			error << "ALSA-MCD: poll reports error."
			      << endmsg;
			return -1;
		}

		if (pfd.revents == 0) {
			// timed out, such as when the device is paused
			continue;
		}

		xrun_detected = false;

		if ((capture_frames_avail = snd_pcm_avail_update (capture_handle)) < 0) {
			if (capture_frames_avail == -EPIPE) {
				xrun_detected = true;
			} else {
				error << "unknown ALSA avail_update return value ("
				      << capture_frames_avail
				      << ')'
				      << endmsg;
				return -1;
			}
		}

		if ((playback_frames_avail = snd_pcm_avail_update (playback_handle)) < 0) {
			if (playback_frames_avail == -EPIPE) {
				xrun_detected = true;
			} else {
				error << "unknown ALSA avail_update return value ("
				      << capture_frames_avail
				      << ')'
				      << endmsg;
				return -1;
			}
		}
		
		if (xrun_detected) {
			if (xrun_recovery ()) {
				return -1;
			}
			continue;
		}

		/* Process 1 fragment at a time */

		avail = capture_frames_avail;
		while ((size_t) avail >= _frames_per_cycle) {

			AudioThing *new_audio_thing = 0;
			
			/* Pop all pending audiothing connection requests
			   of the queue. i.e. just use the most recent one.
			*/
			
			while (audiothing_requests.pop (new_audio_thing) == 0) 
				;
			
			if (new_audio_thing) {
				connect_to_audiothing (new_audio_thing);
			}
			
			channels_not_done = channel_done_bits;

			if (audiothing) {
				
				/* if we are going to fold multiple tracks
				   from the audiothing onto fewer physical
				   output channels, then silence the playback
				   fragment for each output channel. This way,
				   the additive operations in merge_memcpy*()
				   will do the right thing.
				*/
				
				prefold_all_channels ();
				
				if (audiothing->process (_frames_per_cycle, _bytes_per_cycle)) {
					error << "ALSA::MCD: audiothing processing error - stopping."
					      << endmsg;
					return -1;
				}
			} 

			if ((last_mask != input_monitor_mask) && !all_monitor_in) {
				hw->set_input_monitor_mask (input_monitor_mask);
				last_mask = input_monitor_mask;
			} 
			
			/* if we've been asked to do "bypass" monitoring, and
			   we're using h/w that doesn't support this, copy
			   all input channels to output channels, paying no
			   attention to anything done by `audiothing'
			*/
			
			if (!_hw_monitoring && all_monitor_in) {
				copy_all_channels ();
			}
			
			if (channels_not_done) {
				process_untouched_channels ();
			}

#ifdef ALSA_MCD_WITH_GAIN
			compute_gains ();
#endif ALSA_MCD_WITH_GAIN
			compute_levels ();

			if (audiothing) {
				if (audiothing->post_process ()) {
					error << "ALSA::MCD: audiothing post-process error - halting"
					      << endmsg;
					return -1;
				}
			}

			avail -= _frames_per_cycle;
			capture_frag++;
			capture_frag %= _nfragments;
			playback_frag++;
			playback_frag %= _nfragments;
		
			snd_pcm_mmap_forward (capture_handle, _frames_per_cycle);
			snd_pcm_mmap_forward (playback_handle, _frames_per_cycle);
		}

		if (avail) {
//			printf ("going back to poll with %d left avail\n", avail);
		}
	}

	return 0;
}

int
MultiChannelDevice::xrun_recovery ()

{
	size_t chn;
	ssize_t capture_delay;
	ssize_t playback_delay;
	ssize_t max_delay;
	int err;

	for (chn = 0; chn < _nchannels; chn++) {
		complete_silence_on_channel (chn);
	}

	if ((err = snd_pcm_delay (capture_handle, &capture_delay))) {
		fatal << "ALSA::MCD: cannot determine capture delay ("
		      << snd_strerror (err)
		      << ')'
		      << endmsg;
		/*NOTREACHED*/
	}

	if ((err = snd_pcm_delay (playback_handle, &playback_delay))) {
		fatal << "ALSA::MCD: cannot determine playback delay ("
		      << snd_strerror (err)
		      << ')'
		      << endmsg;
		/*NOTREACHED*/
	}

	max_delay = capture_delay > playback_delay ? capture_delay : playback_delay;

	if (max_delay <= 0) {
		fatal << "ALSA::MCD: xrun report, but max_delay is "
		      << max_delay
		      << endmsg;
		/*NOTREACHED*/
		return 0;
	}

	if (audiothing && audiothing->xrun_recoverable (max_delay) == false) {
		/* don't report an error here, its distracting */
		return -1;
	} 

	/* XXX this is problematic. we need to position the ptrs to 
	   align to the frag size. FIX ME.
	*/

	if (capture_delay > 0) {
		cout << "ALSA::MCD: capture xrun of "
		     << capture_delay
		     << " frames ("
		     << ((float) capture_delay / (float) _frame_rate) * 1000.0f
		     << "msecs)"
		     << endl;

		if (snd_pcm_mmap_forward (capture_handle, capture_delay) != capture_delay) {
			return -1;
		}
	}

	if (playback_delay < 0) {
		
		/* the playback delay ranges from -inf to 0 */

		playback_delay = -playback_delay;

		cout << "ALSA::MCD: playback xrun of "
		     << playback_delay
		     << " frames ("
		     << ((float) playback_delay / (float) _frame_rate) * 1000.0f
		     << "msecs)"
		     << endl;

		if (snd_pcm_mmap_forward (playback_handle, playback_delay) != playback_delay) {
			return -1;
		}
	}

	return 0;
}	

void
MultiChannelDevice::copy_all_channels ()

{
	for (size_t n = 0; n < _nchannels; n++) {
		true_copy[n] (playback_addr[n][playback_frag],
			      capture_addr[n][capture_frag],
			      _channel_fragment_bytes,
			      interleave_skip);
		mark_channel_done (n);
	}
}

void
MultiChannelDevice::process_untouched_channels ()

{
	for (size_t n = 0; n < _nchannels; n++) {
		if ((channels_not_done & (1<<n))) { 
			if (input_monitor_mask & (1<<n)) {
				copy_channel (n, n, _channel_fragment_bytes,0);
			} else {
				if (!silent[n][playback_frag]) {
					silence_on_channel (n);
					silent[n][playback_frag] = 1;
				}
			}
		}
	}
}

void
MultiChannelDevice::compute_gains ()

{
	for (size_t chn = 0; chn < _nchannels; chn++) {

		/* we can't do this if we are using h/w monitoring for
		   this channel, and we shouldn't waste time doing it
		   if it has not been requested either (gain == 1)
		*/
		   
		
		if ((_hw_monitoring && (input_monitor_mask & (1<<chn))) ||
		    gain[chn] == 1) {
			continue;
		}
		
		switch (_sample_format) {
		case SND_PCM_SFMT_S32_LE:
			adjust_gain<int> ((int *) playback_addr[chn][capture_frag],
					  _frames_per_cycle,
					  interleaved ? _nchannels : 1,
					  gain[chn]);
			break;
		case SND_PCM_SFMT_S16_LE:
			adjust_gain<short> ((short *) playback_addr[chn][capture_frag],
					    _frames_per_cycle,
					    interleaved ? _nchannels : 1,
					    gain[chn]);
		}
	}
}

void
MultiChannelDevice::compute_levels ()

{
	size_t chn;

	for (chn = 0; chn < _nchannels; chn++) {
		switch (_sample_format) {
		case SND_PCM_SFMT_S32_LE:
			if ((input_monitor_mask & (1<<chn)) ||
			    (input_meter_mask & (1<<chn))) {
				_peak_input_power[chn] = 0;
				compute_peak<int> ((int *) capture_addr[chn][capture_frag],
						   &_peak_input_power[chn], 
						   _frames_per_cycle,
						   interleaved ? _nchannels : 1,
						   max_level);
			}
			if (output_meter_mask & (1<<chn)) {
				_peak_output_power[chn] = 0;
				compute_peak<int> ((int *) playback_addr[chn][capture_frag],
						   &_peak_output_power[chn], 
						   _frames_per_cycle,
						   interleaved ? _nchannels : 1,
						   max_level);
			}
			compute_overs<int> (chn, max_level, min_level);
			break;

		case SND_PCM_SFMT_S16_LE:
			if ((input_monitor_mask & (1<<chn)) ||
			    (input_meter_mask & (1<<chn))) {
				_peak_input_power[chn] = 0;
				compute_peak<short> ((short *) capture_addr[chn][capture_frag],
						     &_peak_input_power[chn], 
						     _frames_per_cycle,
						     interleaved ? _nchannels : 1,
						     max_level);
			}
			if (output_meter_mask & (1<<chn)) {
				_peak_output_power[chn] = 0;
				compute_peak<short> ((short *) playback_addr[chn][capture_frag],
						     &_peak_output_power[chn], 
						     _frames_per_cycle,
						     interleaved ? _nchannels : 1,
						     max_level);
			}
			compute_overs<short> (chn, max_level, min_level);
			break;
		}
	}
}


template<class SampleType> void
MultiChannelDevice::compute_overs (size_t chn, 
				   SampleType maximum,
				   SampleType minimum)

{
	size_t n;
	SampleType *sample;
	size_t overlen;

	sample = (SampleType *) playback_addr[chn][playback_frag];
	overlen = over_length[chn];
	
	for (n = 0; n < _frames_per_cycle; n++) {
		if (*sample == maximum || *sample == minimum) {
			overlen++;
		} else if (overlen) {
			if (overlen > over_length_short) {
				overs[ShortOver][chn]++;
			}
			if (overlen > over_length_long) {
				overs[LongOver][chn]++;
			}
			overlen = 0;
		}
		if (interleaved) {
			sample += _nchannels;
		} else {
			sample++;
		}
	}
	
	/* post-loop check the final status of overlen */
	
	if (overlen) {
		if (overlen > over_length_short) {
			overs[ShortOver][chn]++;
		}
		if (overlen > over_length_long) {
			overs[LongOver][chn]++;
		}
	}
	
	over_length[chn] = overlen;
}

void
MultiChannelDevice::reset_overs (size_t chn)

{
	overs[ShortOver][chn] = 0;
	overs[LongOver][chn] = 0;
	over_length[chn] = 0;
}

void
MultiChannelDevice::reset_all_overs ()

{
	memset (over_length, 0, sizeof (size_t) * _nchannels);
	memset (overs[ShortOver], 0, sizeof (size_t) * _nchannels);
	memset (overs[LongOver], 0, sizeof (size_t) * _nchannels);
}

void
MultiChannelDevice::reset_peak (size_t chn)

{
	_peak_input_power[chn] = 0;
	_peak_output_power[chn] = 0;
}

void
MultiChannelDevice::reset_all_peaks ()

{
	memset (_peak_input_power, 0, sizeof (float) * _nchannels);
	memset (_peak_output_power, 0, sizeof (float) * _nchannels);
}

void
MultiChannelDevice::reset_gain ()

{
	for (size_t n = 0; n < _nchannels; n++) {
		gain[n] = 1;
	}
}

int
MultiChannelDevice::audio_start ()

{
	int err;

	if ((err = snd_pcm_prepare (playback_handle)) < 0) {
		error << "ALSA-HW: prepare error for playback on "
		      << alsa_name
		      << " ("
		      << snd_strerror(err)
		      << ')'
		      << endmsg;
		return -1;
	}

	if (independent_go) {
		if ((err = snd_pcm_prepare (capture_handle)) < 0) {
			error << "ALSA-HW: prepare error for capture on "
			      << alsa_name
			      << " ("
			      << snd_strerror(err)
			      << ')'
			      << endmsg;
			return -1;
		}
	}

	if (_hw_monitoring) {
		hw->set_input_monitor_mask (input_monitor_mask);
	}

	/* fill playback buffer with zeroes, and mark 
	   all fragments as having data.
	*/
	
	for (size_t chn = 0; chn < _nchannels; chn++) {
		complete_silence_on_channel (chn);
	}

	snd_pcm_mmap_forward (playback_handle, _nfragments * _frames_per_cycle);

	if ((err = snd_pcm_start (playback_handle)) < 0) {
		error << "could not start playback ("
		      << snd_strerror (err)
		      << ')'
		      << endmsg;
		return -1;
	}

	if (independent_go) {
		if ((err = snd_pcm_start (capture_handle)) < 0) {
			error << "could not start capture ("
			      << snd_strerror (err)
			      << ')'
			      << endmsg;
			return -1;
		}
	}

	return 0;
}

int
MultiChannelDevice::audio_stop ()

{
	int err;

	if ((err = snd_pcm_drop (playback_handle)) < 0) {
		error << "ALSA::MCD: channel flush for playback failed ("
		      << snd_strerror (err)
		      << ')'
		      << endmsg;
		return -1;
	}

	if (independent_go) {
		if ((err = snd_pcm_drop (capture_handle)) < 0) {
			error << "ALSA::MCD: channel flush for capture failed ("
			      << snd_strerror (err)
			      << ')'
			      << endmsg;
			return -1;
		}
	}
	return 0;
}

void *
MultiChannelDevice::do_work ()

{
	pthread_mutex_lock (&restart_lock);

	while (!work_no_more()) {
		audio_start ();
		if (io_loop ()) {
			audio_stop ();
			halted ();
			pthread_cond_wait (&_restart, &restart_lock);
		}
	}
	pthread_mutex_unlock (&restart_lock);
	return 0;
}

void
MultiChannelDevice::restart ()

{
	pthread_mutex_lock (&restart_lock);
	pthread_cond_signal (&_restart);
	pthread_mutex_unlock (&restart_lock);
}

void
MultiChannelDevice::connect_to_audiothing (AudioThing *thing)

{
	if (audiothing) {
		disconnect_from_audiothing (audiothing);
	}

	audiothing = thing;

	if (thing) {
		thing->going_away.connect 
			(slot 
			 (*this, 
			  &MultiChannelDevice::disconnect_from_audiothing));
		
		thing->post_connect ();

		for (size_t n = 0; n < _nchannels; n++) {
			setup_io_function_pointers (n);
		}
	}
}

void
MultiChannelDevice::disconnect_from_audiothing (AudioThing *t)

{
	audiothing = 0;
}

void
MultiChannelDevice::setup_io_function_pointers (size_t chn)

{
	if (audiothing == 0) {
		/* we'll do it when we connect */
		return;
	}

	switch (_sample_bytes) {
	case 2:
		if (folding[chn]) {
			if (interleaved) {
				channel_copy[chn] = merge_memcpy_interleave_d16_s16;
			} else {
				channel_copy[chn] = merge_memcpy_d16_s16;
			}
		} else {
			if (interleaved) {
				channel_copy[chn] = memcpy_interleave_d16_s16;
			} else {
				channel_copy[chn] = memcpy_fake;
			}
		}

		if (interleaved) {
			true_copy[chn] = memcpy_interleave_d16_s16;
		} else {
			true_copy[chn] = memcpy_fake;
		}

		read_via_copy[chn] = sample_move_dS_s16;
		write_via_copy[chn] = folding[chn] ? sample_merge_d16_sS : sample_move_d16_sS;
		break;

	case 4:
		if (folding[chn]) {
			if (interleaved) {
				channel_copy[chn] = merge_memcpy_interleave_d32_s32;
			} else {
				channel_copy[chn] = merge_memcpy_d32_s32;
			}
		} else {
			if (interleaved) {
				channel_copy[chn] = memcpy_interleave_d32_s32;
			} else {
				channel_copy[chn] = memcpy_fake;
			}
		}

		if (interleaved) {
			true_copy[chn] = memcpy_interleave_d32_s32;
		} else {
			true_copy[chn] = memcpy_fake;
		}

		read_via_copy[chn] = sample_move_dS_s32u24;
		write_via_copy[chn] = folding[chn] ? sample_merge_d32u24_sS : sample_move_d32u24_sS;
		break;
	}
}

void
MultiChannelDevice::xfade_in_from_channel_s32 (size_t chn, ARDOUR::Sample *buf, 
					       size_t nsamples, size_t offset,
					       size_t xfade_position)
{
	int *newptr;
	ARDOUR::Sample s;

	newptr = (int *)(capture_addr[chn][capture_frag]) + offset;

	while (nsamples--) {

		s = ((*newptr >> 8) / SAMPLE_MAX_24BIT) * xfade_in[xfade_position];
		*buf++ = (*buf * xfade_out[xfade_position]) + s;
		newptr++;
		
		if (xfade_position == 0) {
			break;
		}

		xfade_position--;
	}

	if (nsamples) {
		read_via_copy[chn] (buf,
				    (char *) newptr, 
				    nsamples,
				    interleave_skip);
	}
	
	mark_channel_done (chn);
	return;
}

void
MultiChannelDevice::xfade_in_from_channel_s16 (size_t chn, ARDOUR::Sample *buf, 
					       size_t nsamples, size_t offset,
					       size_t xfade_position)
{
	short *newptr;
	ARDOUR::Sample s;

	newptr = (short *)(capture_addr[chn][capture_frag]) + offset;

	while (nsamples--) {

		s = (*newptr / SAMPLE_MAX_16BIT) * xfade_in[xfade_position];
		*buf++ = (*buf * xfade_out[xfade_position]) + s;
		newptr++;
		
		if (xfade_position == 0) {
			break;
		}

		xfade_position--;
	}

	if (nsamples) {
		read_via_copy[chn] (buf,
				    (char *) newptr, 
				    nsamples,
				    interleave_skip);
	}
	
	mark_channel_done (chn);
	return;
}

void
MultiChannelDevice::xfade_out_from_channel_s32 (size_t chn, ARDOUR::Sample *buf, 
						size_t nsamples, size_t offset,
						size_t xfade_position)
{
	int *newptr;
	ARDOUR::Sample s;

	newptr = (int *)(capture_addr[chn][capture_frag]) + offset;

	while (nsamples--) {

		/* Switching curves doesn't work, because we end up
		   with the steepest curve at the end of the fade out.
		   So instead, read them backwards.
		*/

		s = ((*newptr >> 8) / SAMPLE_MAX_24BIT) * xfade_in[_xfade_samples - xfade_position - 1];
		*buf++ = (*buf * xfade_out[_xfade_samples - xfade_position - 1]) + s;
		newptr++;
		
		if (xfade_position == 0) {
			break;
		}

		xfade_position--;
	}
	mark_channel_done (chn);
}

void
MultiChannelDevice::xfade_out_from_channel_s16 (size_t chn, ARDOUR::Sample *buf, 
						size_t nsamples, size_t offset,
						size_t xfade_position)
{
	short *newptr;
	ARDOUR::Sample s;

	newptr = (short *)(capture_addr[chn][capture_frag]) + offset;

	while (nsamples--) {

		/* Switching curves doesn't work, because we end up
		   with the steepest curve at the end of the fade out.
		   So instead, read them backwards.
		*/

		s = (*newptr / SAMPLE_MAX_16BIT) * xfade_in[_xfade_samples - xfade_position - 1];
		*buf++ = (*buf * xfade_out[_xfade_samples - xfade_position - 1]) + s;
		newptr++;
		
		if (xfade_position == 0) {
			break;
		}

		xfade_position--;
	}
	mark_channel_done (chn);
}

void
MultiChannelDevice::xfade_out_from_channel (size_t chn, 
					    ARDOUR::Sample *buf, 
					    size_t nsamples, size_t offset,
					    size_t xfade_position)
{
	switch (_sample_format) {
	case SND_PCM_SFMT_S16_LE:
		xfade_out_from_channel_s16 (chn, buf, nsamples, offset, xfade_position);
		break;

	case SND_PCM_SFMT_S32_LE:
		xfade_out_from_channel_s32 (chn, buf, nsamples, offset, xfade_position);
		break;

	default:
		fatal << "programming error: bad sample format for xfade_in" << endmsg;
		/*NOTREACHED*/
	}
}

void
MultiChannelDevice::xfade_in_from_channel (size_t chn, 
					   ARDOUR::Sample *buf, 
					   size_t nsamples, size_t offset,
					   size_t xfade_position)
{
	switch (_sample_format) {
	case SND_PCM_SFMT_S16_LE:
		xfade_in_from_channel_s16 (chn, buf, nsamples, offset, xfade_position);
		break;

	case SND_PCM_SFMT_S32_LE:
		xfade_in_from_channel_s32 (chn, buf, nsamples, offset, xfade_position);
		break;

	default:
		fatal << "programming error: bad sample format for xfade_in" << endmsg;
		/*NOTREACHED*/
	}
}

void
MultiChannelDevice::silence_all_channels ()

{
	size_t chn;

	for (chn = 0; chn < _nchannels; chn++) {
		silence_on_channel (chn);
	}
}

void
MultiChannelDevice::request_all_monitor_input (bool yn)

{
	if (_hw_monitoring) {
		if (yn) {
			hw->set_input_monitor_mask (~0U);
		} else {
			hw->set_input_monitor_mask (input_monitor_mask);
		}
	}

	all_monitor_in = yn;
}

void
MultiChannelDevice::request_meter_input (size_t chn, bool yn)

{
	if (yn) {
		input_meter_mask |= (1<<chn);
	} else {
		input_meter_mask &= ~(1<<chn);
	}
}

void
MultiChannelDevice::request_meter_output (size_t chn, bool yn)

{
	if (yn) {
		output_meter_mask |= (1<<chn);
	} else {
		output_meter_mask &= ~(1<<chn);
	}
}

void
MultiChannelDevice::request_monitor_input (size_t chn, bool yn)

{
	if (yn) {
		if (input_monitor_mask & (1<<chn)) {
			return;
		}
		input_monitor_mask |= (1<<chn);
	} else {
		if (!(input_monitor_mask & (1<<chn))) {
			return;
		}
		input_monitor_mask &= ~(1<<chn);
	}

	if (!_hw_monitoring && !yn) {
		complete_silence_on_channel (chn);
	}

	/* We pick up the new mask during ::process(), when, if the mask
	   is different than the last value we set their, we'll reset
	   it.
	*/

	/* Tell anyone who cares about the state of input monitoring */

	input_monitor_change (chn, yn);
}

void
MultiChannelDevice::set_hw_monitoring (bool yn)

{
	if (yn) {
		_hw_monitoring = true;

		if (all_monitor_in) {
			hw->set_input_monitor_mask (~0U);
		} else {
			hw->set_input_monitor_mask (input_monitor_mask);
		}

	} else {
		_hw_monitoring = false;
		hw->set_input_monitor_mask (0);
	}
}


void
MultiChannelDevice::compute_xfade ()

{
	size_t n;
	double dn;
	double fract;
	double maxlog;
	double divisor;

	if (xfade_out) {
		delete xfade_out;
	}
	if (xfade_in) {
		delete xfade_in;
	}

	xfade_in = new float[_xfade_samples];	
	xfade_out = new float[_xfade_samples];

	/* Special case zero fade duration */

	if (_xfade_samples == 0) {
		for (size_t n = 0; n < _xfade_samples; n++) {
			xfade_in[n] = 1;
			xfade_out = 0;
		}
		return;
	}

	divisor = _xfade_samples;
	maxlog = log10 (1.0 / divisor);

	for (n = 0, dn = 1.0; n < _xfade_samples; n++, dn += 1.0) {
		fract = dn / divisor;

		/* xfade_in will contain values going from 1.0 to 0.0
		   xfade_out will contain values going from 0.0 to 1.0

		   This may appear to be the wrong ordering, but we
		   do it like this because we actually access them
		   in reverse index order: the first values used
		   are xfade_{in,out}[_xfade_samples-1] and the last
		   ones are xfade_{in,out}[0].
		*/

		xfade_in[n] = log10(fract) / maxlog;
		xfade_out[n] = 1 - xfade_in[n];
	}
	
}

void
MultiChannelDevice::set_xfade_length (size_t msecs)

{
	_xfade_samples = (size_t) ceil (((double) msecs / 1000.0) * (double) _frame_rate);
	compute_xfade ();
}

int
MultiChannelDevice::check_capabilities ()

{
	return 0;
}

int
MultiChannelDevice::change_sample_clock (MultiChannelHardware::SampleClockMode mode)

{
	return hw->change_sample_clock (mode);
}

void
MultiChannelDevice::complete_silence_on_channel (size_t chn)

{
	size_t frag;

	for (frag = 0; frag < _nfragments; frag++) {
		if (interleaved) {
			memset_interleave 
				(playback_addr[chn][frag], 0, 
				 _channel_fragment_bytes,
				 interleave_unit,
				 interleave_skip);
		} else {
			memset (playback_addr[chn][frag], 0, 
				_channel_fragment_bytes);
		}
	}

	/* of course, we've actually done more than necessary here,
	   since we set the entire buffer to zero. however, it remains
	   true that this channel has been handled for the current
	   cycle.
	*/
	
	mark_channel_done (chn);
}

void
MultiChannelDevice::adjust_out_channels (int old_chn, int new_chn)
{
	if (old_chn >= 0) {
		if (out_cnt[old_chn] == 2) {
			folding[old_chn] = false;
			setup_io_function_pointers (old_chn);
		} else if (out_cnt[old_chn] == 1) {
			if (!(input_monitor_mask & (1<<old_chn))) {
				complete_silence_on_channel (old_chn);
			} 
		}
		out_cnt[old_chn]--;
	}
	
	if (new_chn >= 0) {
		out_cnt[new_chn]++;
		if (out_cnt[new_chn] > 1) {
			folding[new_chn] = true;
			setup_io_function_pointers (new_chn);
		}
	}
}

void
MultiChannelDevice::adjust_in_channels (int old_chn, int new_chn)
{
	if (old_chn >= 0) {
		in_cnt[old_chn]--;
		if (in_cnt[old_chn] == 0) {
			request_monitor_input (old_chn, false);
		}
	}
	
	if (new_chn >= 0) {
		in_cnt[new_chn]++;
	}
}

void
MultiChannelDevice::prefold_all_channels ()

{
	for (size_t chn = 0; chn < _nchannels; chn++) {
		if (folding[chn]) {
			silence_on_channel (chn);
		}
	}
}

void
MultiChannelDevice::set_over_lengths (size_t longval, size_t shortval)
{
	over_length_long = longval;
	over_length_short = shortval;
}
