/*
    Copyright (C) 1999-2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: alsa_device.h,v 1.50 2000/11/04 04:53:33 pbd Exp $
*/

#ifndef __alsa_device_h__
#define __alsa_device_h__

#include <string.h>
#include <sys/types.h>

#include <sys/asoundlib.h>
#include <sigc++/signal_system.h>
#include <pbd/thread.h>
#include <pbd/lock_free_fifo.h>

#include "alsa_multichannel_hw.h"
#include "memops.h"

class AudioThing;

namespace ALSA {
	
class MultiChannelDevice : public QMThread, public SigC::Object {
  public:
	MultiChannelDevice (const string &name,
			    size_t rate,
			    size_t frames_per_cycle,
			    bool real_time);

	~MultiChannelDevice ();

	bool ok()                 { return _ok; }

	size_t sample_bytes()     { return _sample_bytes; }
	size_t frame_rate ()      { return _frame_rate; }
	int    sample_format ()   { return _sample_format; }
	size_t nchannels ()       { return _nchannels; }
	size_t bytes_per_cycle()  { return _bytes_per_cycle; }
	size_t frames_per_cycle() { return _frames_per_cycle; }

	int set_parameters (size_t frames_per_cycle,
			    size_t sample_rate);
	
	void  restart ();

	void set_over_lengths (size_t longval, size_t shortval);

	void connect_to_audiothing (AudioThing *t);

	/* These three functions are called by a Track
	   to move data to/from the audio h/w.

	   The first two perform simple byte-for-byte
	   copies from the audio h/w buffer into
	   the buffer provided as an argument.

	   The third+fourth use a pre-computed cross-fade
	   curve to mix the data from the audio h/w
	   into the buffer provided, which is assumed
	   to hold existing audio data.
	*/

	void read_from_channel (size_t channel, ARDOUR::Sample *buf, 
				size_t nsamples, size_t offset) {
		read_via_copy[channel]
			(buf, 
			 capture_addr[channel][capture_frag] + offset,
			 nsamples, 
			 interleave_skip);
	}

	void write_to_channel (size_t channel, ARDOUR::Sample *buf,
			       size_t nsamples, size_t offset) {
		if (!all_monitor_in) {
			write_via_copy[channel] 
				(playback_addr[channel][playback_frag] + offset,
				 buf, 
				 nsamples, 
				 interleave_skip);
		}
		mark_channel_done (channel);
	}

	/* We used to do these with templates, but that doesn't work
	   when the fundamental math that needs to be done is
	   different for S32 and S16.
	*/

        void xfade_in_from_channel_s32 (size_t chn, ARDOUR::Sample *buf, 
					size_t nsamples, size_t offset,
					size_t xfade_position);
        void xfade_in_from_channel_s16 (size_t chn, ARDOUR::Sample *buf, 
					size_t nsamples, size_t offset,
					size_t xfade_position);

	void xfade_out_from_channel_s32 (size_t chn, ARDOUR::Sample *buf, 
					 size_t nsamples, size_t offset,
					 size_t xfade_position);
	void xfade_out_from_channel_s16 (size_t chn, ARDOUR::Sample *buf, 
					 size_t nsamples, size_t offset,
					 size_t xfade_position);
		
	void xfade_in_from_channel (size_t channel, ARDOUR::Sample *buf,
				    size_t samples, size_t offset,
				    size_t xfade_sample);

	void xfade_out_from_channel (size_t channel, ARDOUR::Sample *buf,
				     size_t samples, size_t offset,
				     size_t xfade_sample);
	
	/* This is used to monitor input */

	void copy_channel (size_t input_channel, 
			   size_t output_channel,
			   size_t nsamples,
			   size_t offset) {
		if (!hw_monitoring() && !all_monitor_in) {
			channel_copy[input_channel] 
				(playback_addr[output_channel][playback_frag] + offset,
				 capture_addr[input_channel][capture_frag] + offset,
				 nsamples * _sample_bytes,
				 interleave_skip);
		}
		mark_channel_done (output_channel);
	}
	
	/* Note that these functions, like the previous ones, operate
	   ONLY on the current fragment.  
	*/
	
	void silence_on_channel (size_t chn) {
		if (interleaved) {
			memset_interleave 
				(playback_addr[chn][playback_frag],
				 0, _bytes_per_cycle,
				 interleave_unit,
				 interleave_skip);
		} else {
			memset (playback_addr[chn][playback_frag],
				0, _bytes_per_cycle);
		}
		mark_channel_done (chn);
	}
	void silence_all_channels ();

	/* This completely silences a channel, by zero-filling
	   EVERY fragment. This is a very important difference
	   with the related functions above.
	*/

	void complete_silence_on_channel (size_t chn);

	/* This does the right thing for a channel depending
	   on whether we are monitoring input or not. The "not"
	   case is handled by generating silence.
	*/

	void monitor_channel (size_t chn);

	void reset_overs (size_t chn);
	void reset_all_overs ();
	void reset_peak (size_t chn);
	void reset_all_peaks ();
	void reset_gain ();

	void set_gain (size_t chn, float level) {
		gain[chn] = level;
	}

	float peak_output_power (size_t chn) {
		return _peak_output_power[chn];
	}

	float peak_input_power (size_t chn) {
		return _peak_input_power[chn];
	}

	bool has_hw_monitoring() {
		return hw->capabilities() & 
			MultiChannelHardware::Cap_HardwareMonitoring;
	}

	bool has_clock_lock_reporting() {
		return hw->capabilities() & 
			MultiChannelHardware::Cap_ClockLockReporting;
	}

	void set_hw_monitoring (bool yn);
	bool hw_monitoring() {
		return _hw_monitoring;
	}

	vector<MultiChannelHardware::ChannelDescription> &stereo_descriptors () {
		return hw->stereo_descriptors ();
	}

	void request_meter_input (size_t chn, bool yn);
	void request_meter_output (size_t chn, bool yn);

	void request_all_monitor_input (bool yn);
	void request_monitor_input (size_t chn, bool yn);

	void toggle_monitor_input (size_t chn) {
		request_monitor_input (chn, !(input_monitor_mask & (1<<chn)));
	}

	bool monitor_in (size_t chn) {
		return all_monitor_in || (input_monitor_mask & (1<<chn));
	}

	enum OverType {
		ShortOver = 0,
		LongOver = 1
	};
	
	size_t over_count (size_t chn, OverType ot) {
		return overs[ot][chn];
	}

	enum ChannelStatus {
		Lock = 0x1,
		NoLock = 0x2,
		Sync = 0x4,
		NoSync = 0x8
	};

	SigC::Signal2<void,size_t,char> channel_status_change;
	SigC::Signal2<void,size_t,bool> input_monitor_change;
	SigC::Signal0<void> halted;

	/* Cross-fade related stuff */

	size_t xfade_samples () { return _xfade_samples; }
	void   set_xfade_length (size_t msecs);

	int change_sample_clock (MultiChannelHardware::SampleClockMode);

	void adjust_out_channels (int old_chn, int new_chn);
	void adjust_in_channels (int old_chn, int new_chn);

  private:
	AudioThing *audiothing;
	LockFreeFIFO<AudioThing *> audiothing_requests;

	pthread_cond_t  _restart;
	pthread_mutex_t restart_lock;

	bool      _ok;
	bool       first_select;
	bool       interleaved;
	size_t     interleave_unit;
	size_t     interleave_skip;
	size_t     src_interleave_skip;
	string     alsa_name;
	int        card_type;
	size_t    _nchannels;
	size_t    _frame_rate;
	size_t    _sample_bytes;
	bool       *folding;
	size_t     *out_cnt;
	size_t     *in_cnt;
	bool      _hw_monitoring;

	size_t    _frames_per_cycle;
#define _fragment_frames _frames_per_cycle
	size_t    _bytes_per_cycle; /* per channel */
#define _channel_fragment_bytes _bytes_per_cycle
	size_t     fragment_bytes;

	int       _sample_format;
	float     _max_sample_val;
	size_t    _nfragments;
	size_t     expected_capture_frame;
	size_t     _xfade_samples;
	float      *xfade_in;
	float      *xfade_out;
	int        max_level;
	int        min_level;

	size_t    *overs[2];
	size_t    *over_length;
	float    *_peak_input_power;
	float    *_peak_output_power;
	float     *gain;
	size_t     capture_frag;
	size_t     playback_frag;
	size_t     capture_boundary;
	size_t     playback_boundary;

	char      *channel_refs;

	bool         all_monitor_in;

	size_t channels_not_done;
	size_t channel_done_bits;
	size_t input_monitor_mask;
	size_t last_mask;
	size_t input_meter_mask;
	size_t output_meter_mask;
	bool **silent;
	bool   independent_go;
	bool   noninterleaved;
	
	size_t over_length_short;
	size_t over_length_long;

	snd_ctl_t              *ctl_handle;
	snd_pcm_t              *playback_handle;
	snd_pcm_t              *capture_handle;
	char                 ***playback_addr;
	char                 ***capture_addr;
	ssize_t                 capture_frames_avail;
	ssize_t                 playback_frames_avail;

	int configure (int alsa_channel);

	void compute_levels ();
	template<class SampleType> void compute_overs (size_t chn,
						       SampleType max,
						       SampleType min);
	void compute_xfade ();
	void compute_gains ();

	void disconnect_from_audiothing (AudioThing *t);

	void *do_work ();
	int   io_loop ();
	static void *start_thread (void *);

	int  configure_switches ();
	int  check_card_type ();
	int  check_capabilities ();
	
	int   audio_start ();
	int   audio_stop ();

	MultiChannelHardware *hw;  
	int hw_specific ();

	int hammerfall_hardware ();
	int generic_hardware ();
	
	int xrun_recovery ();
	
	void get_fragment_numbers ();
	void wrapup ();
	void process_untouched_channels ();
	void prefold_all_channels ();
	void copy_all_channels ();

	void mark_channel_done (size_t chn) {
		channels_not_done &= ~(1<<chn);
		silent[chn][playback_frag] = false;
	}

	void setup_io_function_pointers (size_t chn);

	typedef void (*ReadCopyFunction)  (ARDOUR::Sample *dst, char *src,
					     size_t src_bytes,
					     size_t src_skip_bytes);
	typedef void (*WriteCopyFunction) (char *dst, ARDOUR::Sample *src,
					     size_t src_bytes,
					     size_t dst_skip_bytes);
	typedef void (*CopyCopyFunction)  (char *dst, char *src,
					     size_t src_bytes,
					     size_t skip_bytes);

	/* each channel gets a function for read/write/copy */

	ReadCopyFunction *read_via_copy;
	WriteCopyFunction *write_via_copy;
	CopyCopyFunction *channel_copy;
	CopyCopyFunction *true_copy;
};

}; /* namespace ALSA */

#endif __alsa_device_h__
