/*
    Copyright (C) 1999 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: ardour_ui.cc,v 1.58 2000/11/04 04:53:33 pbd Exp $
*/


#include <algobase.h>

#include <math.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>

#include <iostream>

#include <gtk--.h>
#include <pbd/stl_delete.h>
#include <pbd/error.h>
#include <pbd/basename.h>
#include <pbd/pathscanner.h>
#include <gtkmmext/pix.h>
#include <gtkmmext/pixmap_button.h>
#include <gtkmmext/utils.h>
#include <gtkmmext/led.h>
#include <gtkmmext/click_box.h>

#include <midi++/port.h>
#include <midi++/mmc.h>

#include "ardour.h"
#include "ardour_ui.h"
#include "track_ui.h"
#include "audio_clock_ui.h"
#include "alsa_device.h"
#include "location_ui.h"

using namespace ARDOUR;
using namespace Gtkmmext;
using namespace Gtk;

ARDOUR_UI *ARDOUR_UI::theHdrUI = 0;

gint
ARDOUR_UI::update_tracks (void *arg)

{
	ARDOUR_UI *ui = (ARDOUR_UI *) arg;
	size_t n;

	for (n = 0; n < Config->max_tracks; n++) {
		ui->track_ui[n]->update ();
	}

	return TRUE;
}

ARDOUR_UI::ARDOUR_UI (ALSA::MultiChannelDevice *mcdp, 
		      MIDI::MachineControl *mmcp,
		      int *argc, char ***argv, 
		      const char *rc_file)

	: Gtkmmext::UI ("ARDOUR", argc, argv, rc_file),
	  mcd (mcdp),
	  mmc (mmcp),

	  /* button table */
	  button_table (15, 3),

	  /* adjuster table */

	  adjuster_table (3, 3),

	  /* adjustments */

	  xfade_adjustment (5, 0, 1500, 1, 50, 0),
	  preroll_adjustment (0, 0, 5, 0.1, 1, 0),
	  postroll_adjustment (0, 0, 5, 0.1, 1, 0),


	  /* new transport dialog */
	  new_tape_dialog (GTK_WINDOW_DIALOG),
	  new_tape_length_adjustment (20, 1, 100, 1, 10),
	  new_tape_length (new_tape_length_adjustment),
	  new_tape_tracks_adjustment (Config->max_tracks, 1, 
				      Config->max_tracks, 1, 4),
	  new_tape_tracks (new_tape_tracks_adjustment),
	  new_tape_tracks_label ("Tracks"),
	  new_tape_name_label ("ID"), 
	  new_tape_length_label ("Length (mins)"),
	  new_tape_ok ("OK"),
	  new_tape_cancel ("Cancel"),

	  wipe_window (GTK_WINDOW_DIALOG),
	  wipe_level (0)


{
	/* XXX how to check that Gtk::UI worked OK ? */

	if (theHdrUI == 0) {
		theHdrUI = this;
	}

	if (setup_windows (*argc, *argv)) {
		return;
	}

	starting.connect (slot (*this, &ARDOUR_UI::startup));
	stopping.connect (slot (*this, &ARDOUR_UI::shutdown));

	/* start with timecode, metering enabled
	*/

	global_meter_button->set_active (true);
	clock_button->set_active (true);
	
	if (mcd->hw_monitoring ()) {
		hw_monitor_button->set_active (true);
	} 

	meter_timer_tag = -1;
	blink_handler_tag = -1;
	
	start_metering ();

	/* until we get a tape, certain buttons are insensitive.
	*/

	single_button_record_button->set_sensitive (false);

	mcd->channel_status_change.connect
		(slot (this, &ARDOUR_UI::channel_status_redisplay));
	mcd->halted.connect
		(slot (this, &ARDOUR_UI::mcd_halted));

	/* Add whatever port the MMC object is using to the input
	   loop
	*/

	if (mmc) {
		MIDI::Port *port = &(mmc->port());
		gdk_input_add (port->selectable(), GDK_INPUT_READ,
			       (void (*) (void *, gint, GdkInputCondition))
			       MIDI::Port::gtk_read_callback,
			       port);
	}

	tape = 0;
	tape_selector_popup = 0;
	last_key_press_time = 0;
	
	/* See comments in ardour_ui.h to understand what this is
	   all about.
	*/

	fakeTransportDisplay.ui = this;
	fakeRecordDisplay.ui = this;
	fakeHalt.ui = this;

	_ok = true;
}

ARDOUR_UI::~ARDOUR_UI ()

{

}

void
ARDOUR_UI::channel_status_redisplay (size_t channel, char status)

{
	if (channel >= Config->max_tracks) {
		return;
	}

	if (status & ALSA::MultiChannelDevice::Lock) {
		if (status & ALSA::MultiChannelDevice::Sync) {
			track_ui[channel]->show_lock_and_sync ();
		} else {
			track_ui[channel]->show_lock ();
		}
	} else {
		track_ui[channel]->show_no_lock ();
	}
}

void
ARDOUR_UI::start_metering ()

{
	if (meter_timer_tag <= 0) {
		meter_timer_tag = gtk_timeout_add (100, update_tracks, this);
		metering = true;
	}
}

void
ARDOUR_UI::stop_metering ()

{
	if (meter_timer_tag > 0) {
		gtk_timeout_remove (meter_timer_tag);
		meter_timer_tag = -1;
	}
	metering = false;
}

void
ARDOUR_UI::toggle_metering ()

{
	if (global_meter_button->get_active()) {
		/* enable */
		if (!metering) {
			start_metering ();
		}
	} else {
		if (metering) {
			stop_metering ();
			clear_meters ();
		}
	}
}

void
ARDOUR_UI::toggle_clock ()

{
	if (clock_button->get_active()) {
		clock_ui->run_clock ();
	} else {
		clock_ui->stop_clock ();
	}
}

int
ARDOUR_UI::_load_default_tape (void *arg)

{
	/* this is a oneshot idle function */

	((ARDOUR_UI *) arg)->load_default_tape ();
	return FALSE; /* don't call me again */
}

void
ARDOUR_UI::load_default_tape ()

{
	PathScanner scanner;
	const string *path;
	string regex;

	regex = "^";
	regex += Config->default_tape;
	regex += ".ardour$";

	path = scanner.find_first (Config->tape_path, regex, false, true);

	if (path == 0) {
		error << "Default tape \""
		      << Config->default_tape
		      << "\" not found"
		      << endmsg;
		return;
	} 

	load_tape (Config->default_tape, *path);
	delete path;
}

void
ARDOUR_UI::startup ()

{
	mcd->run ();

	if (Config->default_tape != "") {
		idle_add (_load_default_tape, this);
	}
}

void
ARDOUR_UI::shutdown ()

{
	/* Don't try this at home. If the mcd is what
	   sent a fatal message, then the audio thread
	   is in the middle of sigsuspend(). It will
	   never return. So, look, we're quitting anyway,
	   and everything will get cleaned up when we
	   _exit()
	*/

//	mcd->stop ();
}

void
ARDOUR_UI::map_tape_description ()

{
	char buf[256];
	float sr;
	int sr_precision;

	sr = (float) tape->frame_rate() / 1000.0f;

	switch (tape->frame_rate()) {
	case 48000:
	case 96000:
		sr_precision = 0;
		break;

	case 44100:
	case 88200:
	default:
		sr_precision = 1;
		break;
	}

	sprintf (buf, "%s: %d mins, %d tracks, %.*f kHz", 
		 tape->name().c_str(), 
		 tape->minutes(),
		 tape->ntracks(),
		 sr_precision,
		 sr);

	tape_id_display.set_text (buf);
}

void
ARDOUR_UI::waiting ()

{
	static GdkCursor *clock_cursor = 0;
	GdkWindow *gw;

	if (!clock_cursor) {
		clock_cursor = gdk_cursor_new (GDK_WATCH);
	}

	gdk_window_set_cursor (main_window.get_window(), clock_cursor);

	if (new_tape_dialog.is_realized() &&
	    (gw = new_tape_dialog.get_window()) != 0) {
		gdk_window_set_cursor (gw, clock_cursor);
	}

	flush_pending ();
}

void
ARDOUR_UI::done_waiting ()

{
	GdkWindow *gw;

	gdk_window_set_cursor (main_window.get_window(), 0);

	if (new_tape_dialog.is_realized() &&
	    (gw = new_tape_dialog.get_window()) != 0) {
		gdk_window_set_cursor (gw, 0);
	}

	flush_pending ();
}

int
ARDOUR_UI::build_track_uis ()

{
	size_t n;

	track_ui = new Track_UI * [Config->max_tracks];

	for (n = 0; n < Config->max_tracks; n++) {
		track_ui[n] = new Track_UI (mcd, n);

		/* Pack the various parts of the track UI into
		   display 
		*/

#define packing_factor 35

		meter_base.put
			(track_ui[n]->meter_box(), 
			 (n * packing_factor) + 
			 (track_ui[n]->meter_width()/2), 0);
		upper_track_box.put
			(track_ui[n]->above_box(), n * packing_factor, 0);
		lower_track_box.put
			(track_ui[n]->below_box(), n * packing_factor, 0);
			 
	}

	return 0;
}	

int
ARDOUR_UI::setup_pixmaps ()

{
	int retval = -1;
	
	info << "UI: setting up images used for interface ..." << endmsg;

	if ((rec_enable_buttons = get_pix (Config->pixmap_path, 
				       Config->rec_enable_pix_regexp)) == 0) {
		error << "No on/off button pixmaps found to match " 
		      << Config->rec_enable_pix_regexp
		      << endmsg;
		goto headback;
	}

	rec_enable_buttons->ref ();

	if ((monitor_buttons = get_pix (Config->pixmap_path, 
				       Config->monitor_pix_regexp)) == 0) {
		error << "No monitor button pixmaps found to match " 
		      << Config->monitor_pix_regexp
		      << endmsg;
		goto headback;
	}

	monitor_buttons->ref ();

	if ((toggle_buttons = get_pix (Config->pixmap_path, 
				       Config->toggle_button_pix_regexp)) == 0) {
		error << "No toggle button pixmaps found to match " 
		      << Config->toggle_button_pix_regexp
		      << endmsg;
		goto headback;
	}

	toggle_buttons->ref ();
	
	if ((push_buttons = get_pix (Config->pixmap_path, 
				     Config->push_button_pix_regexp)) == 0) {
		error << "No push button pixmaps found to match " 
		      << Config->push_button_pix_regexp
		      << endmsg;
		goto headback;
	}

	push_buttons->ref ();

	if ((small_push_buttons = get_pix (Config->pixmap_path, 
					   Config->small_push_button_pix_regexp)) == 0) {
		error << "No small push button pixmaps found to match " 
		      << Config->small_push_button_pix_regexp
		      << endmsg;
		goto headback;
	}

	small_push_buttons->ref ();

	if (LED::get_pix (Config->pixmap_path)) {
		error << "No LED pixmaps found in "
		      << Config->pixmap_path
		      << endmsg;
		goto headback;
	}

	retval = 0;

  headback:
	return retval;
}

#define ReferenceAndReturn(x) \
           Pix *ARDOUR_UI::x ## _pix () { x ## s->ref(); return x ## s; }

ReferenceAndReturn(rec_enable_button)
ReferenceAndReturn(monitor_button)
ReferenceAndReturn(push_button)
ReferenceAndReturn(small_push_button)
ReferenceAndReturn(toggle_button)

void
ARDOUR_UI::display_transport_state ()

{
	switch (tape->transport_state()) {
	case Tape::Roll:
	case Tape::Rewind:
	case Tape::Forward:
		xfade_button->set_sensitive (false);
		break;
	case Tape::Stopped:
	case Tape::Locating:
		xfade_button->set_sensitive (true);
		break;
	}
}

void
ARDOUR_UI::display_record_state ()

{
	return;
}

void
ARDOUR_UI::toggle_auto_play ()

{
	toggle_some_tape_state (auto_play_button,
				&Tape::get_auto_play,
				&Tape::set_auto_play);
}

void
ARDOUR_UI::toggle_auto_input ()

{
	toggle_some_tape_state (auto_input_button,
				&Tape::get_auto_input,
				&Tape::set_auto_input);
	
	for (size_t n = 0; n < Config->max_tracks; n++) {
		track_ui[n]->clear_meter ();
	}
}

void
ARDOUR_UI::toggle_dynamic_punch ()

{
	toggle_some_tape_state (dynamic_punch_button,
				&Tape::get_dynamic_punch,
				&Tape::set_dynamic_punch);
}

void
ARDOUR_UI::toggle_rehearse ()

{
	toggle_some_tape_state (rehearse_button,
				&Tape::get_rehearse,
				&Tape::set_rehearse);
}

void
ARDOUR_UI::toggle_auto_loop ()

{
	toggle_some_tape_state (auto_loop_button,
				&Tape::get_auto_loop,
				&Tape::set_auto_loop);
}

void
ARDOUR_UI::toggle_auto_punch ()

{
	toggle_some_tape_state (auto_punch_button,
				&Tape::get_auto_punch,
				&Tape::set_auto_punch);
}

void
ARDOUR_UI::toggle_single_button_record ()

{
	toggle_some_tape_state (single_button_record_button,
				&Tape::get_record_one_touch,
				&Tape::set_record_one_touch);
}

void
ARDOUR_UI::toggle_all_safe ()

{
	toggle_some_tape_state (all_safe_button,
				&Tape::get_all_safe,
				&Tape::set_all_safe);
}

void
ARDOUR_UI::toggle_monitor_in ()
	
{
}

void
ARDOUR_UI::toggle_hw_monitoring ()

{
	mcd->set_hw_monitoring (!mcd->hw_monitoring());
}

void
ARDOUR_UI::toggle_bypass ()

{
	if (bypass_button->get_active ()) {
		mcd->request_all_monitor_input (true);
	} else {
		mcd->request_all_monitor_input (false);
	}
}

void
ARDOUR_UI::map_button_state ()

{
	map_some_tape_state (dynamic_punch_button,
			     &Tape::get_dynamic_punch);
	map_some_tape_state (auto_play_button,
			     &Tape::get_auto_play);
	map_some_tape_state (auto_punch_button,
			     &Tape::get_auto_punch);
	map_some_tape_state (all_safe_button,
			     &Tape::get_all_safe);
	map_some_tape_state (single_button_record_button,
			     &Tape::get_record_one_touch);
	map_some_tape_state (rehearse_button,
			     &Tape::get_rehearse);
	map_some_tape_state (auto_input_button,
			     &Tape::get_auto_input);
	map_some_tape_state (auto_loop_button,
			     &Tape::get_auto_loop);
}

void
ARDOUR_UI::clear_meters ()

{
	for (size_t n = 0; n < Config->max_tracks; n++) {
		track_ui[n]->clear_meter ();
	}
	reset_meters_button->set_active (false);
}

gint
ARDOUR_UI::_blink_handler (void *arg)

{
	((ARDOUR_UI *) arg)->blink_handler ();
	return TRUE;
}

void
ARDOUR_UI::blink_handler ()

{
	if (tape == 0) {
		return;
	}

	for (size_t n = 0; n < Config->max_tracks; n++) {
		track_ui[n]->record_enable_blink (rec_enable_blink_on);
	}

	rec_enable_blink_on = !rec_enable_blink_on;
}

void
ARDOUR_UI::set_xfade ()

{
	size_t xfade_msecs;

	if (tape == 0) {
		return;
	}

	xfade_msecs = (size_t) xfade_adjustment.get_value ();	
	tape->set_xfade_length (xfade_msecs);
}

void
ARDOUR_UI::control_methods_adjusted ()

{
	int which_method;

	which_method = (int) online_control_button->adjustment.get_value();
	switch (which_method) {
	case 0:
		allow_mmc_and_local ();
		break;
	case 1:
		allow_mmc_only ();
		break;
	case 2:
		allow_local_only ();
		break;
	default:
		fatal << "programming error: impossible control method"
		      << endmsg;
	}
}
	

void
ARDOUR_UI::mmc_device_id_adjusted ()

{
	int dev_id = (int)
		mmc_id_button->adjustment.get_value();

	mmc->set_device_id (dev_id);
}

void
ARDOUR_UI::sample_clock_mode_adjusted ()

{
	int clock_mode = (int)
		sample_clock_mode_button->adjustment.get_value();

	switch (clock_mode) {
	case 0:
		mcd->change_sample_clock (ALSA::MultiChannelHardware::AutoSync);
		break;

	case 1:
		mcd->change_sample_clock (ALSA::MultiChannelHardware::WordClock);
		break;

	case 2:
		mcd->change_sample_clock (ALSA::MultiChannelHardware::ClockMaster);
		break;

	default:
		fatal << "programming error: impossible sample clock mode"
		      << endmsg;
	}
}


void
ARDOUR_UI::notebook_switched (Gtk::Notebook_Helpers::Page *page, guint pagenum)

{
	switch (pagenum) {
	case 0: /* ardour recorder */
		switched_to_recorder ();
		break;
	case 1:
		switched_to_editor ();
		break;
	default:
		fatal << "programmer error: unhandled notebook page " 
		      << pagenum
		      << endmsg;
		/*NOTREACHED*/
	}
	last_notebook_page = pagenum;
}

void
ARDOUR_UI::transport_change ()

{
	touch_display (&fakeTransportDisplay);
}

void
ARDOUR_UI::record_status_change ()

{
	touch_display (&fakeRecordDisplay);
}

void
ARDOUR_UI::map_some_tape_state (PixmapButton *button,
				bool (Tape::*get)())
	
{
	if (tape == 0) {
		return;
	}
	
	button->set_active ((tape->*get)());
}

void
ARDOUR_UI::map_control_change (Tape::ControlType t)

{
	switch (t) {
	case Tape::AutoPlay:
		map_some_tape_state (auto_play_button, &Tape::get_auto_play);
		break;

	case Tape::AutoLoop:
		map_some_tape_state (auto_loop_button, &Tape::get_auto_loop);
		break;

	case Tape::Rehearse:
		map_some_tape_state (rehearse_button, &Tape::get_rehearse);
		break;

	case Tape::AllSafe:
		map_some_tape_state (all_safe_button, &Tape::get_all_safe);
		break;

	case Tape::DynamicPunch:
		map_some_tape_state (dynamic_punch_button, &Tape::get_dynamic_punch);
		break;

	case Tape::RecordOneTouch:
		map_some_tape_state (single_button_record_button, &Tape::get_record_one_touch);
		break;

	case Tape::AutoInput:
		map_some_tape_state (auto_input_button, &Tape::get_auto_input);
		break;

	case Tape::AutoPunch:
		map_some_tape_state (auto_punch_button, &Tape::get_auto_punch);
		break;

	case Tape::SendMTC:
//		map_some_tape_state (send_mtc_button, &Tape::get_send_mtc);
		break;

	case Tape::MMCControl:       
//		map_some_tape_state (mmc_control_button, &Tape::get_mmc_control);
		break;
	}
}

void
ARDOUR_UI::toggle_some_tape_state (PixmapButton *button,
				   bool (Tape::*get)(),
				   void (Tape::*set)(bool))

{
	bool button_state;
	bool tape_state;

	if (tape == 0) {
		return;
	}

	button_state = button->get_active ();
	tape_state = (tape->*get)();

	if (button_state != tape_state) {
		(tape->*set) (button_state);
#if 0
	
		/* check that it worked, and reverse
		   the button state if it didn't
		*/

		if ((tape->*get)() != button_state) {
			button->set_active (!button_state);
		}
#endif

	}
}	

void
ARDOUR_UI::unload_tape ()

{
	mcd->connect_to_audiothing (0);

	if (tape) {
		delete tape;
		tape = 0;
	}

	tape_id_display.set_text ("");
	clock_button->set_active (false);

	if (blink_handler_tag > 0) {
		gtk_timeout_remove (blink_handler_tag);
		blink_handler_tag = -1;
	}
}

void
ARDOUR_UI::tape_selected (PopupSelector *popup, 
			  SelectionResult *result)
	
{
	/* This is a bad move: removing the popup means that the
	   buttonRelease event doesn't get delivered to it. This can 
	   cause problems ... Hmmm.
	*/

	popup->hide ();

	if (result->text == 0 || result->text->length() == 0 ||
	    result->data == 0) {
		return;
	}

	load_tape (*result->text, *((string *) result->data));
}

int
ARDOUR_UI::load_tape (const string &name, const string &path)

{
	Tape *new_tape;

	new_tape = new Tape (mcd, mmc, name, path);

	if (!new_tape->ok()) {
		error << "Transport \"" 
		      << name 
		      << "\" did not load successfully"
		      << endmsg;
		delete new_tape;
		return -1;
	}

	unload_tape ();
	connect_to_tape (new_tape);
	return 0;
}

gint
ARDOUR_UI::tape_menu (GdkEventButton *ev)

{
	tape_popup_menu->popup (0, 0, 0, 0);
	return TRUE;
}

static bool 
tape_filter (const string &str)

{
	return (str.length() > 7 && str.find (".ardour") == (str.length() - 7));
}	

static void
available_tape_lister (Gtk::CList &list, void *dirpath)

{
	dirpath_lister (list, (char *) dirpath, tape_filter, false);
}

gint
ARDOUR_UI::new_tape_window_deleted (GdkEventAny *ev)

{
	new_tape_dialog.hide_all ();
	return TRUE;
}

void
ARDOUR_UI::new_tape ()

{
	new_tape_dialog.show_all ();
}

void
ARDOUR_UI::cancel_new_tape ()

{
	new_tape_dialog.hide ();
}

void
ARDOUR_UI::accept_new_tape ()

{
	string name;
	string path;
	size_t ntracks;
	size_t minutes;
	Tape *pending_tape;

	minutes = new_tape_length.get_value_as_int ();
	ntracks = new_tape_tracks.get_value_as_int ();
	name = new_tape_name.get_text();

	if (name.length() == 0) {
		error << "You haven't entered a name for this transport"
		      << endmsg;
		return;
	}

	path = Config->tape_path;
	path += '/';
	path += name;
	path += ".ardour";

	new_tape_dialog.hide ();

	pending_tape = new Tape (mcd,
				 mmc,
				 name,
				 path, 
				 ntracks,
				 minutes,
				 mcd->frame_rate());
	
	while (pending_tape->tracks_still_loading() > 0 ) {
		gtk_main_iteration ();
	}

	if (!pending_tape->ok()) {
		error << "Sorry, I couldn't create the transport as requested"
		      << endmsg;
	} else {
		connect_to_tape (pending_tape);
	}
	pending_tape = 0;
}

void
ARDOUR_UI::open_tape ()

{
	/* popup selector window */

	if (tape_selector_popup == 0) {
		tape_selector_popup = new PopupSelector ("Select Tape",
							 available_tape_lister,
							 (void *) Config->tape_path.c_str());
		tape_selector_popup->set_usize (400, 200);
		tape_selector_popup->selection_made.connect
			(slot (this,&ARDOUR_UI::tape_selected));
	}

	tape_selector_popup->show ();

	/* wait for selection */
}

void
ARDOUR_UI::tape_add_track ()

{
}

void
ARDOUR_UI::tape_extend ()

{
}

void
ARDOUR_UI::transport_locate (size_t new_position)

{
	size_t _preroll;

	if (tape) {
		_preroll = (size_t) floor (mcd->frame_rate() * preroll());

		if (new_position > _preroll) {
			new_position -= _preroll;
		} else {
			new_position = 0;
		}

		tape->request_locate (new_position);
	}
}

gint
ARDOUR_UI::transport_stop (GdkEventButton *ev)

{
	if (tape) {
		tape->request_transport_state (Tape::Stopped);
	}
	return TRUE;
}

gint
ARDOUR_UI::transport_record (GdkEventButton *ev)

{
	if (tape) {
		switch (tape->record_status()) {
		case Tape::Disabled:
			tape->maybe_enable_record ();
			break;
		case Tape::Recording:
		case Tape::Enabled:
			tape->disable_record ();
		}
	}
	return TRUE;
}

gint
ARDOUR_UI::transport_roll (GdkEventButton *ev)

{
	if (tape) {
		tape->request_transport_state (Tape::Roll);
	}
	return TRUE;
}

gint
ARDOUR_UI::transport_rewind (GdkEventButton *ev)

{
	if (tape) {
		tape->use_rf_shuttle_speed ();
		tape->request_transport_state (Tape::Rewind);
	}
	return TRUE;
}

gint
ARDOUR_UI::transport_forward (GdkEventButton *ev)

{
	if (tape) {
		tape->use_rf_shuttle_speed ();
		tape->request_transport_state (Tape::Forward);
	}
	return TRUE;
}

void
ARDOUR_UI::map_transport_state ()

{
	switch (tape->transport_state()) {
	case Tape::Roll:
		transport_rolling ();
		break;
	case Tape::Rewind:
		transport_rewinding ();
		break;
	case Tape::Forward:
		transport_forwarding ();
		break;
	case Tape::Stopped:
		transport_stopped ();
		break;
	case Tape::Locating:
		transport_locating ();
	}
	
	display_transport_state ();
	map_record_state ();
}

void
ARDOUR_UI::map_record_state ()

{
	switch (tape->record_status()) {
	case Tape::Enabled:
		rec_led->red ();
		rec_led->blink ();
		break;
	case Tape::Recording:
		rec_led->red();
		break;
	default:
		rec_led->black ();
	}

	display_record_state ();
}

void
ARDOUR_UI::transport_locating ()

{
	rec_led->black ();
	stop_led->black ();
	roll_led->black();
	rewind_led->black ();
	forward_led->black ();
}

void
ARDOUR_UI::transport_stopped ()

{
	rec_led->black ();
	stop_led->green ();
	roll_led->black();
	rewind_led->black ();
	forward_led->black ();
}

void
ARDOUR_UI::transport_rolling ()

{
	rec_led->no_blink ();

	stop_led->black ();
	roll_led->green ();
	rewind_led->black ();
	forward_led->black ();
}

void
ARDOUR_UI::transport_rewinding ()

{
	rec_led->black ();
	stop_led->black ();
	roll_led->black();
	rewind_led->green ();
	rewind_led->blink ();
	forward_led->black ();
}

void
ARDOUR_UI::transport_forwarding ()

{
	rec_led->black ();
	stop_led->black ();
	roll_led->black();
	rewind_led->black ();
	forward_led->green ();
	forward_led->blink ();
}



void
ARDOUR_UI::allow_local_only ()

{

}

void
ARDOUR_UI::allow_mmc_only ()

{

}

void
ARDOUR_UI::allow_mmc_and_local ()

{

}

void
ARDOUR_UI::reset_meters ()

{
	mcd->reset_all_overs ();
	mcd->reset_all_peaks ();
	clear_meters ();
}

void
ARDOUR_UI::wipe_tape ()

{
	if (tape == 0) {
		return;
	}

	switch (wipe_level) {
	case 0:
		wipe_window.show_all ();
		wipe_window.set_title ("Wipe");
		wipe_commit_label.set_text ("Fill with silence ? (non-destructive)");
		wipe_level++;
		break;

	case 1:
		wipe_window.hide_all ();
		tape->wipe ();
		wipe_level = 0;
	}
}

void
ARDOUR_UI::wipe_cancel ()

{
	wipe_window.hide_all ();
	wipe_level = 0;
}

int
ARDOUR_UI::wipe_delete (GdkEventAny *ev)

{
	wipe_cancel ();
	return TRUE;
}

void
ARDOUR_UI::GlobalClickBox::printer (char buf[32], Gtk::Adjustment &adj, void *arg)

{
	sprintf (buf, "%s", ((GlobalClickBox *) arg)->strings[
		(int) adj.get_value()].c_str());
}

void
ARDOUR_UI::switch_to_notebook_page (int page)

{
	ARDOUR_UI::instance()->notebook.set_page (page);
}

void
ARDOUR_UI::toggle_control_window ()
{
	if (control_window.is_visible()) {
		control_window.hide_all ();
	} else {
		control_window.show_all ();
	}
}

void
ARDOUR_UI::toggle_location_window ()
{
	if (location_window.is_visible()) {
		location_window.hide_all ();
	} else {
		location_window.show_all ();
	}
}

gint
ARDOUR_UI::global_key_release_handler (GdkEventKey *ev)

{
	if (notebook.get_current_page_num() == 1) {
		/* Editor is active, do nothing */
		return TRUE;
	}
	return recorder_key_release (ev);
}

gint
ARDOUR_UI::global_key_press_handler (GdkEventKey *ev)

{
	if (notebook.get_current_page_num() == 1) {
		/* Editor is active, do nothing */
		return TRUE;
	}

	return recorder_key_press (ev);
}

gint
ARDOUR_UI::recorder_key_press (GdkEventKey *ev)

{
	bool set_last_key_time = true;
	bool stopit = false;
	guint32 when = gdk_event_get_time ((GdkEvent *) ev);

	/* ignore keypress events less than 1/4 second apart
	   (GDK time is in milliseconds)
	*/

	if (last_key_press_time && when - last_key_press_time < 250) {
		last_key_press_time = when;
		return TRUE;
	}

	switch (ev->keyval) {
	case GDK_S:
	case GDK_s:
		transport_stop (0);
		break;

	case GDK_P:
	case GDK_p:
		transport_roll (0);
		break;

	case 'L':
	case 'l':
		transport_locate (0);
		break;

	case GDK_space:
		if (tape) {
			if (tape->transport_state() == Tape::Stopped) {
				transport_roll (0);
			} else {
				transport_stop (0);
			}
		}
		break;

	case GDK_R:
	case GDK_r:
		transport_record (0);
		break;

	case GDK_Left:
		stopit = true;
	case GDK_less:
		transport_rewind (0);
		break;

	case GDK_Right:
		stopit = true;
	case GDK_greater:
		transport_forward (0);
		break;

	case GDK_Home:
		break;

	case GDK_End:
		break;

	case GDK_F1:
	case GDK_F2:
	case GDK_F3:
	case GDK_F4:
	case GDK_F5:
	case GDK_F6:
	case GDK_F7:
	case GDK_F8:
	case GDK_F9:
	case GDK_F10:
	case GDK_F11:
	case GDK_F12:
		if (tape) {
			size_t track;

			if (ev->state & (GDK_SHIFT_MASK|GDK_CONTROL_MASK) ==
			    (GDK_SHIFT_MASK|GDK_CONTROL_MASK)) {
				track = ev->keyval - GDK_F1 + 24;
			} else if (ev->state & GDK_SHIFT_MASK) {
				track = ev->keyval - GDK_F1 + 12;
			} else {
				track = ev->keyval - GDK_F1;
			}

			if (ev->state & GDK_MOD2_MASK) {
				tape->toggle_track_monitor_input (track);
			} else {
				tape->toggle_track_record_enable (track);
			}
		}

	default:
		set_last_key_time = false;
		break;
	}

	if (set_last_key_time) {
		last_key_press_time = when;
	}

	if (stopit) {
		gtk_signal_emit_stop_by_name (GTK_OBJECT(transport_base.gtkobj()),
					      "key-press-event");	
	}

	return TRUE;
}

gint
ARDOUR_UI::recorder_key_release (GdkEventKey *ev)

{
	bool stopit = false;

	switch (ev->keyval) {
	case GDK_Left:
		stopit = true;
	case GDK_less:
		transport_stop (0);
		break;

	case GDK_Right:
		stopit = true;
	case GDK_greater:
		transport_stop (0);
		break;
	}

	if (stopit) {
		gtk_signal_emit_stop_by_name (GTK_OBJECT(transport_base.gtkobj()),
					      "key-release-event");	
	}

	return TRUE;
}

void
ARDOUR_UI::mcd_halted ()

{
	touch_display (&fakeHalt);
}

void
ARDOUR_UI::notify_mcd_halted ()

{
	Gtk::Window window (GTK_WINDOW_DIALOG);
	Gtk::Button button ("Restart audio interface");

	window.add (button);
	window.set_title ("ardour: audio interface halt");
	window.show_all ();
	
	while (1) {
		gtk_main_iteration ();
	}
}
