/* gtkeditablewavebuffer.c: Interface for editable wavebuffer.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include "gtkeditablewavebuffer.h"
#include <gtk/gtksignal.h>


static void gtk_editable_wave_buffer_class_init      (GtkEditableWaveBufferClass *klass);
static void gtk_editable_wave_buffer_init            (GtkEditableWaveBuffer      *editable_wave_buffer);

guint modified_signal;
guint insert_data_signal;
guint delete_data_signal;


GtkType
gtk_editable_wave_buffer_get_type (void)
{
  static GtkType editable_wave_buffer_type = 0;

  if (!editable_wave_buffer_type)
    {
      static const GtkTypeInfo editable_wave_buffer_info =
      {
        "GtkEditableWaveBuffer",
        sizeof (GtkEditableWaveBuffer),
        sizeof (GtkEditableWaveBufferClass),
        (GtkClassInitFunc) gtk_editable_wave_buffer_class_init,
        (GtkObjectInitFunc) gtk_editable_wave_buffer_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL
      };

      editable_wave_buffer_type = gtk_type_unique (GTK_TYPE_WAVE_BUFFER, &editable_wave_buffer_info);
    }

  return editable_wave_buffer_type;
}


static void
gtk_editable_wave_buffer_real_destroy (GtkObject *object)
{
  GtkEditableWaveBuffer *editable = GTK_EDITABLE_WAVE_BUFFER (object);

  if (editable->modified)
    {
      gtk_idle_remove (editable->modified_idle);
      editable->modified = FALSE;
    }
}


static void
gtk_editable_wave_buffer_class_init (GtkEditableWaveBufferClass *klass)
{
  GtkObjectClass *object_class;
  GtkWaveBufferClass *wavebuffer_class;

  object_class = (GtkObjectClass *) klass;
  wavebuffer_class = (GtkWaveBufferClass *) klass;

  modified_signal =
    gtk_signal_new ("modified",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (GtkEditableWaveBufferClass, modified),
                    gtk_marshal_NONE__POINTER,
                    GTK_TYPE_NONE,
                    1, GTK_TYPE_POINTER);

  insert_data_signal =
    gtk_signal_new ("insert_data",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (GtkEditableWaveBufferClass, insert_data),
                    gtk_marshal_NONE__POINTER,
                    GTK_TYPE_NONE,
                    1, GTK_TYPE_POINTER);

  delete_data_signal =
    gtk_signal_new ("delete_data",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (GtkEditableWaveBufferClass, delete_data),
                    gtk_marshal_NONE__POINTER,
                    GTK_TYPE_NONE,
                    1, GTK_TYPE_POINTER);

  gtk_object_class_add_signals (object_class, &modified_signal, 1);
  gtk_object_class_add_signals (object_class, &insert_data_signal, 1);
  gtk_object_class_add_signals (object_class, &delete_data_signal, 1);

  object_class->destroy = gtk_editable_wave_buffer_real_destroy;

  klass->modified = NULL;
  klass->insert_data = NULL;
  klass->delete_data = NULL;

  wavebuffer_class->get_datatype = NULL;
  wavebuffer_class->get_length = NULL;
  wavebuffer_class->get_num_channels = NULL;
  wavebuffer_class->get_samples = NULL;
  klass->insert = NULL;
  klass->insert_sequence = NULL;
  klass->replace = NULL;
  klass->remove = NULL;
  klass->undo = NULL;
  klass->redo = NULL;
}


static void
gtk_editable_wave_buffer_init (GtkEditableWaveBuffer *editable)
{
  editable->modified = FALSE;
}


static gint
gtk_editable_wave_buffer_emit_modified (GtkEditableWaveBuffer *editable)
{
  gtk_signal_emit (GTK_OBJECT (editable), modified_signal, &editable->modified_range);
  editable->modified = FALSE;
  return 0;
}


void
gtk_editable_wave_buffer_set_modified (GtkEditableWaveBuffer *editable, 
					guint32 start, guint32 length)
{
  if (editable->modified) 
    {
      g_range_add (&editable->modified_range, start, length);
    }
  else
    {
      g_range_set (&editable->modified_range, start, length);
      editable->modified = TRUE;
      editable->modified_idle =
        gtk_idle_add ((GtkFunction) &gtk_editable_wave_buffer_emit_modified,
                      editable);
    }
}


gint
gtk_editable_wave_buffer_insert (GtkEditableWaveBuffer *editable,
				  guint32 start,
				  piece_t *piece)
{
  return GTK_EDITABLE_WAVE_BUFFER_CLASS (GTK_OBJECT
					  (editable)->klass)->insert (editable, start, piece);
}

gint
gtk_editable_wave_buffer_insert_sequence (GtkEditableWaveBuffer *editable,
					  guint32 start,
					  sequence_t *seq)
{
  return GTK_EDITABLE_WAVE_BUFFER_CLASS (GTK_OBJECT
					  (editable)->klass)->insert_sequence (editable, start, seq);
}

gint
gtk_editable_wave_buffer_replace (GtkEditableWaveBuffer *editable,
				   guint32 start,
				   piece_t *piece)
{
  return GTK_EDITABLE_WAVE_BUFFER_CLASS (GTK_OBJECT
					  (editable)->klass)->replace (editable, start, piece);
}


gint
gtk_editable_wave_buffer_delete (GtkEditableWaveBuffer *editable,
				  guint32 start,
				  guint32 length)
{
  return GTK_EDITABLE_WAVE_BUFFER_CLASS (GTK_OBJECT
					  (editable)->klass)->remove (editable, start, length);
}

gint           
gtk_editable_wave_buffer_undo (GtkEditableWaveBuffer *editable,
			       guint32 depth)

{
  return GTK_EDITABLE_WAVE_BUFFER_CLASS (GTK_OBJECT
					  (editable)->klass)->undo (editable, depth);
}

gint 
gtk_editable_wave_buffer_redo (GtkEditableWaveBuffer *editable,
			       guint32 depth)
{
  return GTK_EDITABLE_WAVE_BUFFER_CLASS (GTK_OBJECT
					  (editable)->klass)->redo (editable, depth);
}
