/*
    Copyright (C) 1998-99 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: precision_timer.cc,v 1.6 1999/11/11 22:34:31 pbd Exp $
*/

#include <math.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif // HAVE_CONFIG_H

#include <sys/time.h>
#include <pbd/precision_timer.h>

#ifdef HAVE_RDTSC

#include <pbd/rdtsc.h>

/* OK, the model here is that we call Time::update() periodically,
   and preferably less than the wrap period for the TSC, and it 
   will update current_time correctly without a system call.
   If we wait too long between updates() (but not way too long,
   since then we'll go through two cycles), we call gettimeofday()
   again to pick up the current time.

   Note, however, that this is very unlikely to happen. Since we use
   all 64 bits of the counter, we get about 1000 years of counter-wrap
   free usage. 
*/

#define _usecs(tv) (((tv)->tv_sec * 1000000) + (tv)->tv_usec)

static unsigned int delay_dummy;
static precision_time_t cycles_per_usec;

PrecisionTimer::PrecisionTimer ()

{
	const unsigned int delay_size = 50000;
	struct timeval start;
	struct timeval stop;
	tsc_t rdtsc_start, rdtsc_stop;
	int rdtsc_cycles;
	int delay_cycles;
	unsigned int gettimeofday_usecs;
	float fcycles_per_usec;
	float fusecs_per_cycle;
	unsigned int delay_time;

	/* This rather bizarre and hard-to-understand stuff is
	   designed to calibrate the cycle counter. The end result is
	   the value of cycles_per_usec. Its very equivalent to
	   the computation of BogoMIPS under Linux, but done
	   here in user space. Of course this makes it less
	   likely to be accurate. But it works for me :)
	*/

	rdtsc_start = rdtsc();
	rdtsc_stop = rdtsc();
	rdtsc_cycles = (int) (rdtsc_stop - rdtsc_start);

	gettimeofday (&start, NULL);
	gettimeofday (&stop, NULL);

	gettimeofday_usecs = _usecs(&stop) - _usecs(&start);

	gettimeofday (&start, NULL);
	rdtsc_start = rdtsc ();
	for (delay_dummy = 0; delay_dummy < delay_size; delay_dummy++);
	gettimeofday (&stop, NULL);
	rdtsc_stop = rdtsc ();

	delay_time = _usecs (&stop) - _usecs (&start);
	delay_cycles = (int) (rdtsc_stop - rdtsc_start);

	fcycles_per_usec = (float) (delay_cycles - rdtsc_cycles) /
	    (float) (delay_time - gettimeofday_usecs);

	fusecs_per_cycle = 1.0 / fcycles_per_usec;

	cycles_per_usec = (int) floor (fcycles_per_usec);
}

precision_time_t 
PrecisionTimer::current ()

{
	return rdtsc() / cycles_per_usec;
}

#else // !HAVE_RDTSC

PrecisionTimer::PrecisionTimer () {}

/* making this static speeds things up just a shade.
 */

static struct timeval __pt_now;

precision_time_t
PrecisionTimer::current ()

{
	gettimeofday (&__pt_now, NULL);

	return (precision_time_t) ((__pt_now.tv_sec * 1000000) + __pt_now.tv_usec);
}

#endif // HAVE_RDTSC
