/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: tape.cc,v 1.60 2000/11/04 04:53:33 pbd Exp $
*/

#include <stdio.h> /* sprintf(3) ... grrr */
#include <math.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <limits.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/time.h>
#include <asm/msr.h>

#include <fstream>

#include <midi++/mmc.h>
#include <midi++/port.h>
#include <pbd/error.h>

#include "configuration.h"
#include "tape.h"
#include "track.h"
#include "riff.h"
#include "alsa_device.h"
#include "utils.h"

using namespace ARDOUR;

Tape::Tape (ALSA::MultiChannelDevice *mcdp,
	    MIDI::MachineControl *mmcp,
	    const string &name,
	    const string &path,
	    size_t track_cnt,
	    size_t mins,
	    size_t frames_per_second)

	: QMThread ("tape butler", start_butler_thread, this,
		    Config->use_real_time, 10),
	  AudioThing ("tape"),
	  pending_events (16)
	  

{
	_ok = false;

	if (!thread_ok()) {
		return;
	}

	first_stage_init (mcdp, mmcp, name, path);

	_ntracks = track_cnt;
	_frames_per_second = frames_per_second;
	_frames = mins * 60 * _frames_per_second;

	if (create()) {
		return;
	}

	if (second_stage_init (true)) {
		return;
	}

	_ok = true;

}

Tape::Tape (ALSA::MultiChannelDevice *mcdp,
	    MIDI::MachineControl *mmcp,
	    const string &name,
	    const string &path)

	: QMThread ("tape butler", start_butler_thread, this, 
		    Config->use_real_time, 10),
	  AudioThing ("tape"),
	  pending_events (16)

{

	_ok = false;

	first_stage_init (mcdp, mmcp, name, path);

	if (read_header()) {
		return;
	}

	if (check_tracks ()){ 
		return;
	}

	if (second_stage_init (false)) {
		return;
	}

	_ok = true;

}

Tape::~Tape ()

{
	size_t n;

	terminate_butler();
	going_away (this);

	if (tracks) {
		for (n = 0; n < _ntracks; n++) {
			if (tracks[n]) {
				delete tracks[n];
			}
		}
		delete tracks;
//		delete track_spaces;
	}
}

int
Tape::create ()

{
	struct stat statbuf;
	
	if (::stat (_path.c_str(), &statbuf) == 0) {
		error << "Tape: \""
		      << AudioThing::name()
		      << "\" already exists. Open it instead."
		      << endmsg;
		return -1;
	}

	if (errno != ENOENT) {
		error << "Tape: \""
		      << AudioThing::name()
		      << "\" might already exist. I'm confused."
		      << endmsg;
		return -1;
	}

	if (mkdir (_path.c_str(), 0755) < 0) {
		error << "Tape: cannot create tape dir \""
		      << _path
		      << "\" ("
		      << strerror (errno)
		      << ")"
		      << endmsg;
		return -1;
	}

	if (write_header ()) {
		error << "Tape: cannot write header file ("
		      << strerror (errno)
		      << endmsg;
		return -1;
	}

	

	return 0;
}

int
Tape::check_tracks ()

{
	char buf[PATH_MAX+1];

	_ntracks = 0;

	while (1) {

		/* Track files are currently sequentially numbered,
		   starting at "1", and we look only for the first
		   "version" or "take" (suffix ".1") of each one.  
		*/

		sprintf (buf, "%s/%d.1", _path.c_str(), _ntracks+1);
		
		if (access (buf, R_OK)) {
			break;
		}

		// XXX check for EDL file, header etc. and others
		
		_ntracks++;
	}


	if (_ntracks == 0) {
		error << "Tape \""
		      << AudioThing::name()
		      << " has no track files."
		      << endmsg;
		return -1;
	}
	
	return 0;
}

void
Tape::first_stage_init (ALSA::MultiChannelDevice *mcdp, 
			MIDI::MachineControl *mmcp,
			const string &name,
			const string &path)

{
	_transport_state = Stopped;
	atomic_set (&_record_status, Disabled);
	_transport_frame = 0;
	_frames = 0;

	pthread_mutex_init (&load_lock, 0);
	pthread_mutex_init (&transport_lock, 0);
	pthread_cond_init (&loaded, 0);

	atomic_set (&_record_status, Disabled);

	_butler_summoned = false;

	mcd = mcdp;
	mmc = mmcp;

	current_shuttle_frames = mcd->frame_rate();
	non_shuttle_frames = Config->rf_samples;

	mtc_enabled = Config->send_midi_timecode;

	if (mmc) {
		_mmc_control = true;
		send_mtc = true;

		mtc_port = &(mmc->port());
		
		mmc->DeferredPlay.connect 
			(slot (*this, &Tape::mmc_deferred_play));
		mmc->Stop.connect 
			(slot (*this, &Tape::mmc_stop));
		mmc->FastForward.connect 
			(slot (*this, &Tape::mmc_fast_forward));
		mmc->Rewind.connect 
			(slot (*this, &Tape::mmc_rewind));
		mmc->RecordStrobe.connect 
			(slot (*this, &Tape::mmc_record_strobe));
		mmc->RecordExit.connect 
			(slot (*this, &Tape::mmc_record_exit));
		mmc->Locate.connect 
			(slot (*this, &Tape::mmc_locate));
		mmc->Shuttle.connect 
			(slot (*this, &Tape::mmc_shuttle));
		mmc->TrackRecordStatusChange.connect
			(slot (*this, &Tape::mmc_record_enable));
		
		/* Set up the qtr frame message */

		mtc_msg[0] = 0xf1;
		mtc_msg[2] = 0xf1;
		mtc_msg[4] = 0xf1;
		mtc_msg[6] = 0xf1;
		mtc_msg[8] = 0xf1;
		mtc_msg[10] = 0xf1;
		mtc_msg[12] = 0xf1;
		mtc_msg[14] = 0xf1;

		/* XXX use correct types for SMPTE 30-drop frame options */
		
		switch ((int) Config->smpte_frames_per_second) {
		case 25:
			mtc_smpte_type = 0x20;
			break;
		case 30:
			mtc_smpte_type = 0x40;
			break;
			
		case 24:
		default:
			mtc_smpte_type = 0;
		};
	} else {
		mtc_port = 0;
		mtc_enabled = 0;
		send_mtc = false;
		_mmc_control = false;
	}

	/* XXX convert path to legal filename */

	AudioThing::_name = name;
	_path = path;

	tracks = 0;
	_ntracks = 0;

	auto_play = false;
	auto_punch = false;
	auto_loop = false;
	auto_looped = false;
	auto_input = false;
	rehearse = false;
	all_safe = false;
	record_one_touch = false;
	dynamic_punch = false;
	stop_pending = false;
	tracks_continuing = 0;
	transport_lock_held = false;
	atomic_set (&_state_changing, 0);
	solo_track = -1; // no solo track

	events = 0;
	next_event = 0;

	last_mtc_smpte_frame = 0;

	/* send initial timecode message */

	send_full_time_code ();

	/* start the tape butler thread */

	run ();
}

int
Tape::second_stage_init (bool new_tracks)

{
	if (load_tracks (new_tracks)) {
		return -1;
	}

	if (_locations.restore (*this)) {
		warning << "Tape: could not restore location information"
			<< endmsg;
	}

	legal_recording_frame = capture_offset;

	return 0;
}

void *
Tape::tape_loader_thread (void *arg)

{
	Tape *t = (Tape *) arg;

	t->_tracks_still_loading = t->really_load_tracks ();
	return 0; /* nobody cares, we're detached */
}

bool
Tape::tracks_still_loading ()

{
	return _tracks_still_loading;
}

int
Tape::really_load_tracks ()

{
	size_t i;
	size_t track_buffer_size;
	size_t nbufs;

	capture_offset = (mcd->frames_per_cycle() + Config->systemic_latency);

	track_buffer_size = (int) floor (mcd->frame_rate() * Config->track_buffer_seconds);
	nbufs = (track_buffer_size * sizeof (ARDOUR::Sample)) / Config->minimum_disk_io_bytes;

	if (nbufs < 3) {
		/* We used to provide a warning here, but it was
		   annoying, so it was dropped.
		*/
		nbufs = 3;
	}

	tracks = new Track * [_ntracks];

	for (i = 0; i < _ntracks; i++) {
		tracks[i] = 0;
	}

	for (i = 0; i < _ntracks; i++) {

		tracks[i] = new Track (*this, mcd, i, nbufs);

		if (!tracks[i]->ok()) {
			error << "Tape: could not create track "
			      << i + 1
			      << endmsg;
			return -1;
		}

		if (i < mcd->nchannels()) {
			tracks[i]->set_input_channel (i);
			tracks[i]->set_output_channel (i);
		} else {
			tracks[i]->set_input_channel 
				(mcd->nchannels() - 1);
			tracks[i]->set_output_channel 
				(mcd->nchannels() - 1);
		}
	}
	return post_track_load ();
}

int
Tape::post_track_load ()

{
	set_auto_input (true);
	return 0;
}

int
Tape::load_tracks (bool new_tracks)

{
	int retval = 0;

	/* New tracks take a while to load, because we typically have
	   to allocate them first. So start up a new thread to do
	   this, and iterate while we do it.  
	*/

	_tracks_still_loading = 1;

	if (new_tracks) {
		pthread_t thr;

		if (pthread_create (&thr, 0, tape_loader_thread, this) != 0) {
			error << "Tape: could not start tape loader thread ("
			      << strerror (errno)
			      << ')'
			      << endmsg;
			return -1;
		}

		if (pthread_detach (thr)) {

			/* I can see no way to recover from this 
			   without leaving a thread around.
			*/

			fatal << "Tape: could not detach tape loader thread ("
			      << strerror (errno)
			      << ')'
			      << endmsg;

			/* NOTREACHED */
		}

		retval = 0;

	} else {
		if ((_tracks_still_loading = really_load_tracks ()) == 0) {
			retval = post_track_load ();
		} else {
			retval = _tracks_still_loading;
		}
	}

	return retval;
}

int
Tape::read_header ()

{
	FILE *f;
	string header_path = _path;
	char field[33];
	float val;
	int ret = 0;

	header_path += "/header";

	if ((f = fopen (header_path.c_str(), "a+")) == 0) {
		error << "Tape \""
		      << _path
		      << "\": no header file"
		      << endmsg;
		return -1;
	}
	
	fseek (f, 0, SEEK_SET);
	while (!feof (f)) {
		/* note the trailing space to gobble newlines */

		if (fscanf (f, "%32s %f ", field, &val) != 2) {
			ret = -1;
			break;
		}

		if (strcmp (field, "rate") == 0) {
			_frames_per_second = (size_t) val;
		} else if (strcmp (field, "frames") == 0) {
			_frames = (size_t) val;
		} else {
			error << "Tape \""
			      << _path
			      << "\": unknown field \""
			      << field
			      << "\" in header."
			      << endmsg;
			ret = -1;
			break;
		}
	}

	fclose (f);
	return ret;
}

int
Tape::write_header ()

{
	FILE *f;
	string header_path = _path;

	header_path += "/header";

	if ((f = fopen (header_path.c_str(), "a+")) == 0) {
		error << "Tape \""
		      << _path
		      << "\": no header file"
		      << endmsg;
		return -1;
	}

	if (fseek (f, 0, SEEK_SET)) {
		fclose (f);
		return -1;
	}

	fprintf (f, "rate %d\n", _frames_per_second);
	fprintf (f, "frames %u\n", _frames);

	if (ferror (f)) {
		fclose (f);
		return -1;
	}

	fclose (f);
	return 0;
}

void
Tape::wait_for_butler ()

{
	pause ();
}

void
Tape::post_transport_motion ()

{
	wait_for_butler ();

	for (size_t n = 0; n < _ntracks; n++) {
		tracks[n]->seek (_transport_frame);
	}
}

void
Tape::request_transport_state (State new_state)

{
	Event *ev = new Event;

	ev->type = Event::TransportState;
	ev->transport_state = new_state;
	ev->action_frame = 0; /* do it right now */
	queue_event (ev);
}

void
Tape::request_locate (size_t target_frame)

{
	Event *ev = new Event;

	ev->type = Event::TransportState;
	ev->transport_state = Locating;
	ev->action_frame = 0; /* do it right now */
	ev->target_frame = target_frame;
	queue_event (ev);
}

void
Tape::locate (size_t frame)

{
	set_transport_position (frame);
	post_transport_motion ();
	send_full_time_code ();
}

void
Tape::stop_transport ()

{
	disable_record ();
	stop_pending = false;
	_transport_state = Stopped;
	tracks_continuing = 0;

	/* flush any pending capture data to disk before we 
	   come to a complete halt.
	*/
	
	for (size_t n = 0; n < _ntracks; n++) {
		tracks[n]->flush_dirty();
	}

	summon_butler ();

	/* when we start recording again, don't record
	   until transport_frame has moved past ...
	*/

	legal_recording_frame = _transport_frame + capture_offset;
			
	audio_unlock_transport ();

	/* Whenever we stop, revert to the regular
	   rewind/forward shuttle speed.
	*/
	
	use_rf_shuttle_speed ();
	send_full_time_code ();

	/* Let any who cares know whats going on */

	TransportStateChange ();
}

void
Tape::audio_lock_transport ()

{
	/* This allows the transport_lock to be a recursive lock
	   without the Pthreads implementation supporting this.
	   
	   This must only be called from the main tape thread
	   or the dynamically created transport threads.
	 */

	if (!transport_lock_held) { 
		pthread_mutex_lock (&transport_lock);
		transport_lock_held = true;
	} 
}

void
Tape::audio_unlock_transport ()

{
	/*
	   This must only be called from the main tape thread
	   or the dynamically created transport threads.
	*/

	pthread_mutex_unlock (&transport_lock);
	transport_lock_held = false;
}


void
Tape::do_set_transport_state_in_another_thread (State new_state,
						size_t locate_frame)
{
	pthread_t thr;

	atomic_set (&_state_changing, 1);
	target_transport_state = new_state;
	target_locate_frame = locate_frame;
	if (pthread_create (&thr, 0, transport_thread, this)) {
		error << "Tape: could not create transport thread"
		      << endmsg;
		atomic_set (&_state_changing, 0);
		return;
	}
	pthread_detach (thr);
}

void *
Tape::transport_thread (void *arg)

{
	((Tape *) arg)->set_transport_state 
		(((Tape *) arg)->target_transport_state);
	atomic_set (&(((Tape *) arg)->_state_changing), 0);
	return 0;
}

void
Tape::set_transport_state (State new_state)

{
	State old_state;

	old_state = transport_state();

	switch (new_state) {
	case Stopped:
	disable_record ();
		switch (_transport_state) {
		case Stopped:
		case Locating:
			/* do nothing */
			break;

		case Roll:
			stop_pending = true;
			break;

		case Rewind:
		case Forward:
			/* prevent post_process() from
			   continuting to advance our position
			   while the butler catches up.
			*/

			stop_transport ();
			post_transport_motion ();

			if (auto_play) {
				_transport_state = Roll;
			} 
			break;
		}
		break;

	case Roll:
		switch (_transport_state) {
		case Forward:
		case Rewind:
			post_transport_motion ();
			send_full_time_code ();
			set_last_roll_location (_transport_frame);
			set_record_status ();
			_transport_state = Roll;
			break;

		case Stopped:
			set_last_roll_location (_transport_frame);
			set_record_status ();
			_transport_state = Roll;
			break;

		case Roll:
		case Locating:
			/* nothing to do */
			break;

		}
		break;

	case Rewind:
	case Forward:
		if (!actively_recording () && transport_state() != Locating) {
			disable_record ();
			_transport_state = new_state;
		}
		break;

	case Locating:
		switch (_transport_state) {
		case Rewind:
		case Forward:
		case Locating:
			/* do nothing */
			break;

		case Roll:
		case Stopped:
			if (!actively_recording()) {
				stop_transport ();
				_transport_state = Locating;
				locate (target_locate_frame);
				if (auto_play || auto_looped) {
					_transport_state = Roll;
				} else {
					_transport_state = Stopped;
					
					/* we need this here */

					TransportStateChange ();
				}
				auto_looped = false;
			}
			break;
		}
	}

	if (_transport_state != Stopped) {
		audio_lock_transport ();
	} 

	if (_transport_state != old_state) {
		TransportStateChange ();
	}
}

void
Tape::set_last_roll_location (size_t where)

{
	Location *l = _locations.last_roll_location();
	
	if (l) {
		l->set_start (where);
	} 
}

void
Tape::set_last_record_location (size_t where)

{
	Location *l = _locations.last_record_location();
	
	if (l) {
		l->set_start (where);
	} 
}

size_t
Tape::adjust_apparent_position (size_t frames)

{
	/* This doesn't actually alter state of the track buffers, but
	   it does do a "tentative" resetting of the current
	   _transport_frame values, on the assumption that we are
	   currently in either Forward or Rewind mode. When we switch
	   to Stopped, we'll call post_transport_motion() to get the
	   buffers serviced.  
	*/

	switch (_transport_state) {
	case Forward:
		increment_transport_position (frames);
		break;
		
	case Rewind:
		decrement_transport_position (frames);
		break;
		
	default:
		fatal << "Tape: bad transport state - programming error" 
		      << endmsg;
		break;
	}
	
	/* !!! WARNING !!!
	   
	   This could potentially take 3 msecs (10 MIDI bytes). It
	   rarely will, however, because there is FIFO in the MIDI
	   device driver, and unless has less than 10 bytes of space,
	   we'll be back in more-or-less no time at all.
	*/

	send_full_time_code ();

	return transport_frame();
}

int
Tape::post_process ()

{
	if (_transport_state != Forward && transport_state() != Rewind) {
		return 0;
	}

	adjust_apparent_position (current_shuttle_frames);

	if (_transport_frame >= _frames || _transport_frame == 0) {
		do_set_transport_state_in_another_thread (Stopped);
	}

	return 0;
}

void
Tape::set_shuttle_speed (size_t frames_per_postroll)

{
	current_shuttle_frames = frames_per_postroll;
}

void
Tape::set_rf_shuttle_speed (size_t frames_per_postroll)

{
	non_shuttle_frames = frames_per_postroll;
}

int
Tape::no_roll (size_t nframes, bool state_change_in_progress)

{
	size_t n;

	switch (_transport_state) {
	case Roll:
		/* must be switching states */
		break;

	case Rewind:
	case Forward:
	case Locating:
		mcd->silence_all_channels();
		break;

	case Stopped:
		for (n = 0; n < _ntracks; n++) {
			if (tracks[n]->no_roll (nframes, state_change_in_progress)) {
				error << "Tape: error in no_roll for track "
				      << n
				      << endmsg;
				return -1;
			}
		}
		break;
	}
	return 0;
}

int
Tape::process (size_t nframes, size_t we_dont_care_about_bytes)

{
	size_t this_nframes;
	size_t end_frame;

	bool recording_is_legal;
	bool recording_is_enabled;
	Event *ev;
	bool tape_needs_butler = false;
	int atomic_solo_track;

	/* if the transport state is changing (c/o another
	   thread), then we can do nothing.
	*/

	if (state_changing()) {
		return no_roll (nframes, true);
	}
	
	/* Process any pending requests */

	if (pending_events.pop (ev) == 0) {
		merge_event (ev);
	}

	/* if we are stopped, then don't do anything */

	if (_transport_state != Roll) {
		return no_roll (nframes, false);
	}

	/* if we are supposed to stop, and no tracks are
	   doing xfade-outs, then stop.
	*/

	if (stop_pending && tracks_continuing == 0) {
		stop_transport ();
		return no_roll (nframes, false);
	}

	/* do not allow solo track status change during an 
	   invocation of ::process.
	*/

	atomic_solo_track = solo_track;

	/* OK, we really are rolling */

	end_frame = transport_frame() + nframes;


	while (nframes) {
		if (next_event) {
			if (next_event->action_frame <= transport_frame()) {
				if (process_event (next_event)) {
					no_roll (nframes, state_changing());
					return 0;
				}
				next_event = next_event->next;
				continue;
			}

			if (next_event->action_frame < end_frame) {
				this_nframes = nframes - (end_frame - next_event->action_frame);
			} else {
				this_nframes = nframes;
			}

		} else {
			this_nframes = nframes;
		}

		/* This needs to be constant as we iterate over all tracks */

		recording_is_enabled = actively_recording ();

		/* recording is OK once the _transport_frame has moved
		   beyond the required offset between playback and
		   capture. if you don't understand that, your name
		   is not Paul Barton-Davis.
		*/

		if ((_transport_frame > legal_recording_frame) && !stop_pending) {
			recording_is_legal = true;
		} else {
			recording_is_legal = false;
		}

		for (size_t n = 0; n < _ntracks; n++) {
			
			int track_status;
			bool track_needs_butler = false;
			bool old_mute = false;

			if (atomic_solo_track >= 0) {
				if (atomic_solo_track == (int) n) {
					old_mute = tracks[n]->set_mute (false);
				} else {
					old_mute = tracks[n]->set_mute (true);
				}
			}
				
			track_status = tracks[n]->roll (this_nframes, 
							recording_is_enabled,
							recording_is_legal,
							track_needs_butler);
			
			if (atomic_solo_track >= 0) {
				tracks[n]->set_mute (old_mute);
			}
			
			if (track_status < 0) {
				
				error << "Tape: error on track "
				      << n+1
				      << " at frame "
				      << transport_frame()
				      << endmsg;
				return -1;

			} else if (track_status > 0) {
				tracks_continuing |= (1<<n);
			} else {
				tracks_continuing &= ~(1<<n);
			}

			if (track_needs_butler) {
				tape_needs_butler = true;
			}
		}

		increment_transport_position (this_nframes);
		nframes -= this_nframes;
		set_next_event ();
	}

	if (_transport_frame >= _frames) {
		stop_transport ();
	} else {
		if (tape_needs_butler) {
			summon_butler ();
		}
	}

	if (mtc_enabled) {
		send_midi_time_code ();
	}

	return 0;
}

int
Tape::toggle_track_monitor_input (size_t track_num)

{
	if (track_num >= _ntracks) {
		return -1;
	}
	
	tracks[track_num]->toggle_monitor_input ();
	return 0;
}

int
Tape::toggle_track_record_enable (size_t track_num)

{
	if (track_num >= _ntracks || all_safe) {
		return -1;
	}

	tracks[track_num]->toggle_record_enable ();
	return 0;
}

int
Tape::set_track_record_enable (size_t track_num, bool yn)

{
	if (record_status() == Recording) {
		if (dynamic_punch == false) {
			return -1;
		}
	}

	if (yn && all_safe) {
		return -1;
	}

	if (track_num < _ntracks) {
		tracks[track_num]->set_record_enable (yn);
		return 0;
	} else {
		return -1;
	}
}

int
Tape::set_track_input_channel (size_t track_num, int chn)

{
	if (track_num < _ntracks) {
		tracks[track_num]->set_input_channel (chn);
		return 0;
	} else {
		return -1;
	}
}

int
Tape::set_track_output_channel (size_t track_num, int chn)

{
	if (track_num < _ntracks) {
		tracks[track_num]->set_output_channel (chn);
		return 0;
	} else {
		return -1;
	}
}

void
Tape::summon_butler ()

{
	_butler_summoned = true;
	poke ();
}

void
Tape::terminate_butler ()

{
	stop ();
}	

void *
Tape::start_butler_thread (void *arg)

{
	return ((Tape *) arg)->main ();
}

#if 0
int
Tape::track_space_compare (const void *ts1, const void *ts2)

{
	/* reverse the sense of the comparison, because we want a list
	   of tracks ordered by which has the *most* space to read
	   data into their playback buffer.
	*/

	return ((TrackSpace *) ts2)->space - 
		((TrackSpace *) ts1)->space;
}
#endif

void *
Tape::do_work ()

{
	int err;
	size_t n;
	int sleep_is_ok = 1;

#ifdef TRACE_TIMING
	unsigned long now, then;
#endif TRACE_TIMING


	if (!_butler_summoned) {
		suspend();
		return 0;
	}

	_butler_summoned = false;
		
#ifdef TRACE_TIMING
	rdtscl (then);
#endif TRACE_TIMING

	for (n = 0; n < _ntracks; n++) {
		
		if ((err = tracks[n]->read_ahead()) > 0) {
			sleep_is_ok = 0;
		} else if (err < 0) {
			error << "Butler read ahead failure on track "
			      << n
			      << endmsg;
			return (void *) -1;
		}
	}
	
	for (n = 0; n < _ntracks; n++) {
		bool foo;

		if ((err = tracks[n]->write_behind (foo)) > 0) {
			sleep_is_ok = 0;
		} else if (err < 0) {
			error << "Butler write-behind failure on track "
			      << n
			      << endmsg;
			return (void *) -1;
		}
	}

	
#ifdef TRACE_TIMING
	rdtscl (now);
	if (now - then > 45000) {
		printf ("BUTLER refill of %.2fMB took %.6f secs %.2fMB/sec\n",
			((float) (24 * Config->minimum_disk_io_bytes)
			 / 1048576.0f),
			(float) (now - then) / 450000000.0f,
			((float) (24 * Config->minimum_disk_io_bytes)
			 / 1048576.0f) / 
			((float) (now - then) / 450000000.0f));
	}
#endif TRACE_TIMING
	
	if (sleep_is_ok) {
		suspend ();  /* this just causes us not to start over again */
	}

	return 0;
}

void
Tape::set_auto_input (bool yn)

{
	auto_input = yn;

	for (size_t n = 0; n < _ntracks; n++) {
		tracks[n]->reset_monitor_input ();
	}

	ControlChanged (AutoInput);
}

void
Tape::auto_punch_start_changed (Location *location)

{
	/* XXX these should be with an atomic replace_event() */
	remove_event (location->start(), Event::PunchIn);
	add_event (location->start(), Event::PunchIn);
}	

void
Tape::auto_punch_end_changed (Location *location)

{
	/* XXX these should be with an atomic replace_event() */
	remove_event (location->end(), Event::PunchOut);
	add_event (location->end(), Event::PunchOut);
}	

void
Tape::set_auto_punch_location (Location *location)

{
	auto_punch_start_changed_connection.disconnect();
	auto_punch_end_changed_connection.disconnect();

	auto_punch_start_changed_connection = location->start_changed.connect
		(slot (this, &Tape::auto_punch_start_changed));

	auto_punch_end_changed_connection = location->end_changed.connect
		(slot (this, &Tape::auto_punch_end_changed));
		
	add_event (location->start(), Event::PunchIn);
	add_event (location->end(), Event::PunchOut);

	location->set_auto_punch (true);
}

void
Tape::remove_auto_punch_location (Location *location)

{
	auto_punch_start_changed_connection.disconnect();
	auto_punch_end_changed_connection.disconnect();

	remove_event (location->start(), Event::PunchIn);
	remove_event (location->end(), Event::PunchOut);

	location->set_auto_punch (false);
}

void
Tape::auto_loop_start_changed (Location *location)

{
	/* XXX these should be with an atomic replace_event() */
	remove_event (0, Event::AutoLoop);
	add_event (location->end(), Event::AutoLoop, location->start());
}	

void
Tape::auto_loop_end_changed (Location *location)

{
	/* XXX these should be with an atomic replace_event() */
	remove_event (0, Event::AutoLoop);
	add_event (location->end(), Event::AutoLoop, location->start());
}	

void
Tape::set_auto_loop_location (Location *location)

{
	auto_loop_start_changed_connection.disconnect();
	auto_loop_end_changed_connection.disconnect();

	auto_loop_start_changed_connection = location->start_changed.connect
		(slot (this, &Tape::auto_loop_start_changed));
	auto_loop_end_changed_connection = location->end_changed.connect
		(slot (this, &Tape::auto_loop_end_changed));

	add_event (location->end(), Event::AutoLoop, location->start());
	location->set_auto_loop (true);
}

void
Tape::remove_auto_loop_location (Location *location)

{
	auto_loop_start_changed_connection.disconnect();
	auto_loop_end_changed_connection.disconnect();
	remove_event (0, Event::AutoLoop);
	location->set_auto_loop (false);
}

void
Tape::delete_auto_locations ()

{
	events = 0;

	/* XXX fix memory leak caused by not deleting events */

	auto_punch_end_changed_connection.disconnect ();
	auto_punch_start_changed_connection.disconnect ();
	auto_loop_start_changed_connection.disconnect ();
	auto_loop_end_changed_connection.disconnect ();
}

void
Tape::add_event (size_t action_frame, Event::Type type, size_t target_frame)

{
	Event *ev = new Event;

	ev->type = type;
	ev->action_frame = action_frame;
	ev->target_frame = target_frame;

	queue_event (ev);
}


void
Tape::remove_event (size_t frame, Event::Type type)

{
	Event *ev = new Event;

	ev->action_frame = frame;
	ev->type = type;
	ev->remove = true;

	queue_event (ev);
}

void
Tape::merge_event (Event *ev)

{
	Event *evp;
	Event *prev;

	if (ev->remove) {
		_remove_event (ev->action_frame, ev->type);
		return;
	}

	/* Handle immediate events right here */

	if (ev->action_frame == 0) {
		process_event (ev);
		delete ev;
		return;
	}

	for (evp = events, prev = 0; evp; prev = evp, evp = evp->next) {
		if (ev->action_frame < evp->action_frame) {
			break;
		}
	}
	
	if (prev) {
		if (ev->type != Event::TransportState) {
			if (ev->action_frame == prev->action_frame &&
			    ev->type == prev->type) {
				error << "Tape: cannot have two events of the same type ("
				      << ev->type
				      << " in this case) at the same time (frame "
				      << ev->action_frame
				      << ')'
				      << endmsg;
				return;
			}
		}
		prev->next = ev;
	} else {
		events = ev;
	}

	ev->next = evp;
	set_next_event ();
}

void
Tape::_remove_event (size_t frame, Event::Type type)

{
	Event *evp;
	Event *prev;

	if (type == Event::AutoLoop) {
		for (evp = events, prev = 0; evp; prev = evp, evp = evp->next) {
			if (evp->type == type) {
				break;
			}
		}
	} else {
		for (evp = events, prev = 0; evp; prev = evp, evp = evp->next) {
			if (evp->action_frame == frame && evp->type == type) {
				break;
			}
		}
	}

	if (evp) {
		if (prev) {
			prev->next = evp->next;
		} else {
			events = events->next;
		}
		delete evp;
	} else {
		// whatever ... the caller was being too careful, or something
	}

	set_next_event ();
}

void
Tape::set_next_event ()

{
	/* if there are no events, there can be no next event */

	if (events == 0) {
		next_event = 0;
		return;
	}

	/* obey immediate events (frame == 0) */

	if (events->action_frame == 0) {
		next_event = events;
		return;
	}

	/* if there is no previously existing next event, then use
	   the first one we have.
	*/

	if (next_event == 0) {
		next_event = events;
	}

	/* now adjust to use the one that happens next */

	while (next_event) {
		if (next_event->action_frame >= transport_frame()) {
			return;
		}
		next_event = next_event->next;
	}
}

int
Tape::process_event (Event *ev)

{
	int retval = 0;

	switch (ev->type) {
	case Event::TransportState:
		do_set_transport_state_in_another_thread (ev->transport_state, ev->target_frame);
		retval = 1; /* XXX but we are never called from that
			       part of roll, so its irrelevant
			    */
		break;
		
	case Event::PunchIn:
		if (auto_punch) {
			if (record_status() == Enabled) {
				enable_record ();
			}
		}
		break;
		
	case Event::PunchOut:
		if (auto_punch) {
			disable_record ();
		}
		break;

	case Event::AutoLoop:
		if (auto_loop) {
			auto_looped = true;
			do_set_transport_state_in_another_thread (Locating, ev->target_frame);
			retval = 1;
		}
		break;

	default:
		fatal << "Programming error: illegal event type in process_event ("
		      << ev->type
		      << ')'
		      << endmsg;
		/*NOTREACHED*/
		break;
	};

	return retval;
}

bool
Tape::set_xfade_length (size_t msecs)

{
	bool ret = false;

	if (record_status() == Recording) {
		warning << "You can't adjust the cross fade length "
		 	   "while recording"
			<< endmsg;
	} else {
		mcd->set_xfade_length (msecs);
		ret = true;
	}

	return ret;
}

void
Tape::mmc_deferred_play (MIDI::MachineControl &mmc)

{
	if (_mmc_control) {
		request_transport_state (Roll);
	}
}

void
Tape::mmc_record_strobe (MIDI::MachineControl &mmc)

{
	if (!_mmc_control) 
		return;

	if (!auto_punch) {
		enable_record ();
	} else {
		atomic_set (&_record_status, Enabled);
		RecordEnabled ();
	}
	request_transport_state (Roll);
}

void
Tape::mmc_record_exit (MIDI::MachineControl &mmc)

{
	if (_mmc_control) {
		disable_record ();
	}
}

void
Tape::mmc_stop (MIDI::MachineControl &mmc)

{
	if (_mmc_control) {
		request_transport_state (Stopped);
	}
}

void
Tape::mmc_rewind (MIDI::MachineControl &mmc)

{
	if (_mmc_control) {
		request_transport_state (Rewind);
	}
}

void
Tape::mmc_fast_forward (MIDI::MachineControl &mmc)

{
	if (_mmc_control) {
		request_transport_state (Forward);
	}
}

void
Tape::mmc_locate (MIDI::MachineControl &mmc,
		  const MIDI::byte *mmc_tc)

{
	if (!_mmc_control) {
		return;
	}

	size_t sr = mcd->frame_rate();
	size_t target_frame;
	size_t fps;

	fps = mmc_tc[0] & 0x30;

	target_frame = ((mmc_tc[0] & 0xf)* 60 * 60 * sr) +
		(mmc_tc[1] * 60 * sr) +
		(mmc_tc[2] * sr);
	

	if (target_frame > _frames) {
		target_frame = _frames;
	}

	request_locate (target_frame);
}

void
Tape::mmc_shuttle (MIDI::MachineControl &mmc, float speed, bool forw)

{
	if (!_mmc_control) {
		return;
	}

	if (Config->shuttle_speed_threshold >= 0 &&
	    speed > Config->shuttle_speed_threshold) {
		speed *= Config->shuttle_speed_factor;
	}

	current_shuttle_frames = (size_t) ((float) speed * mcd->frames_per_cycle());
	request_transport_state (forw ? Forward : Rewind);
}

void
Tape::mmc_record_enable (MIDI::MachineControl &mmc, size_t track, bool enabled)

{
	if (_mmc_control) {
		set_track_record_enable (track, enabled);
	}
}

int
Tape::send_full_time_code ()

{
	size_t hrs;
	size_t mins;
	size_t secs;
	size_t frames;
	MIDI::byte msg[10];

	if (mtc_port == 0 || !send_mtc) {
		return 0;
	}

	msg[0] = 0xf0;
	msg[1] = 0x7f;
	msg[2] = 0x7f;
	msg[3] = 0x1;
	msg[4] = 0x1;
	msg[9] = 0xf7;

	sample_to_smpte_clock_parts (transport_frame(),
				     mcd->frame_rate(),
				     Config->smpte_frames_per_second,
				     &hrs,
				     &mins,
				     &secs,
				     &frames);

	msg[5] = mtc_smpte_type | hrs;
	msg[6] = mins;
	msg[7] = secs;
	msg[8] = frames;
	
	if (mtc_port->midimsg (msg, sizeof (msg))) {
		error << "Tape: could not send full MIDI time code"
		      << endmsg;
		return -1;
	}

	last_mtc_smpte_frame = 
		(size_t) floor ((double) transport_frame() /
				((1.0/Config->smpte_frames_per_second)/
				 (1.0/mcd->frame_rate())));

	return 0;
}

int
Tape::send_midi_time_code ()

{
	size_t hrs;
	size_t mins;
	size_t secs;
	size_t frames;
	MIDI::byte hb;
	size_t elapsed_frames;
	size_t current_smpte_frame;
	size_t n;

	if (mtc_port == 0 || !send_mtc)  {
		return 0;
	}

	current_smpte_frame = 
		(size_t) floor ((double) transport_frame() /
				((1.0/Config->smpte_frames_per_second)/
				 (1.0/mcd->frame_rate())));

	elapsed_frames = current_smpte_frame - last_mtc_smpte_frame;

	/* Assumption: elapsed frames < Config->smpte_frames_per_second

	   i.e. there is never more than 1 second between the
	   last_mtc_smpte_frame and right now. This means that we only
	   have to compute the clock parts once, then we just
	   increment the frames part each time.
	*/

	sample_to_smpte_clock_parts (transport_frame(),
				     mcd->frame_rate(),
				     Config->smpte_frames_per_second,
				     &hrs,
				     &mins,
				     &secs,
				     &frames);

	for (n = 0; n < elapsed_frames; n++) {
		
		hb = mtc_smpte_type | hrs;
		
		mtc_msg[1] =  0x00 | (frames & 0xf);
		mtc_msg[3] =  0x10 | ((frames & 0xf0) >> 4);
		mtc_msg[5] =  0x20 | (secs & 0xf);
		mtc_msg[7] =  0x30 | ((secs & 0xf0) >> 4);
		mtc_msg[9] =  0x40 | (mins & 0xf);
		mtc_msg[11] = 0x50 | ((mins & 0xf0) >> 4);
		mtc_msg[13] = 0x60 | (hb & 0xf);
		mtc_msg[15] = 0x70 | ((hb & 0xf0) >> 4);

		if (mtc_port->midimsg (mtc_msg, sizeof (mtc_msg))) {
			error << "Tape: cannot send quarter-frame MTC message ("
			      << strerror (errno)
			      << ')'
			      << endmsg;
			return -1;
		}

		frames++;
	}

	last_mtc_smpte_frame += elapsed_frames;
	return 0;
}

void
Tape::maybe_enable_record ()

{
	atomic_set (&_record_status, Enabled);

	if (_transport_state == Roll) {
		if (!auto_punch) {
			enable_record ();
		} 
	} else {
		if (record_one_touch) {
			request_transport_state (Roll);
		} 
		RecordEnabled ();
	}

}

void
Tape::set_record_status ()

{
	switch (record_status()) {
	case Enabled:
		if (!auto_punch) {
			enable_record ();
		}
		break;

	case Recording:
		disable_record ();
		break;

	default:
		break;
	}
}

int 
Tape::wipe ()

{
	size_t n;
	int err = 0;

	if (pthread_mutex_trylock (&transport_lock)) {
		warning << "Tape is running - cannot wipe"
			<< endmsg;
		return -1;
	}
		
	if (_transport_state != Stopped) {
		warning << "Tape is not stopped - cannot wipe"
			<< endmsg;
		err = -1;
		goto out;
	}
	
	for (n = 0; n < _ntracks; n++) {
		if (wipe_track (n, false)) {
			err = -1;
			break;
		}
	}

  out:
	pthread_mutex_unlock (&transport_lock);
	return err;
}

int 
Tape::wipe_track (size_t trk, bool take_lock)

{
	int err = 0;

	if (take_lock) {
		pthread_mutex_lock (&transport_lock);
	}

	if (_transport_state != Stopped) {
		warning << "Tape is not stopped - cannot wipe track"
			<< endmsg;
		err = -1;
		goto out;
	}
	
	if (tracks[trk]->wipe ()) {
		error << "Tape: could not wipe track "
		      << trk
		      << endmsg;
		err = -1;
	}

	post_transport_motion ();

  out:
	if (take_lock) {
		pthread_mutex_unlock (&transport_lock);
	}
	

	return err;
}

void
Tape::post_connect ()

{
	for (size_t n = 0; n < _ntracks; n++) {
		tracks[n]->set_monitor_input (false);
	}
}

int
Tape::save_locations ()

{
	return _locations.save (*this);
}

int
Tape::restore_locations ()

{
	return _locations.restore (*this);
}

sequence_t **
Tape::sequences ()

{
	sequence_t **ss;
	size_t n;

	ss = new sequence_t * [_ntracks];

	for (n = 0; n < _ntracks; n++) {
		ss[n] = tracks[n]->sequence();
	}

	return ss;
}

void
Tape::set_track_solo (size_t tracknum, bool yn)

{
	if (yn) {

		solo_stack.push (tracknum);
		solo_track = tracknum;

	} else if (!solo_stack.empty()) {

		solo_stack.pop();

		if (!solo_stack.empty()) {
			solo_track = solo_stack.top();
		} else {
			solo_track = -1;
		}
	}
}

void
Tape::set_track_mute (size_t tracknum, bool yn)

{
	tracks[tracknum]->set_mute (yn);
}

bool
Tape::xrun_recoverable (ssize_t frames)

{
	if (_transport_state == Roll) {
		return false;
	} 
	return true;
}
