/*
    Copyright (C) 2000 Paul Barton-Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: trackview.cc,v 1.7 2000/11/03 19:45:39 pbd Exp $
*/

#include <stdio.h> // for sprintf, grrr 
#include <stdlib.h>
#include <math.h>

#include <string>

#include <gtkwaveform-0.2/gtkedlwavebuffer.h>
#include <gtkwaveform-0.2/gtkwaveview.h>

#include <pbd/error.h>
#include <pbd/touchable.h>

#include <gtkmmext/utils.h>
#include <gtkmmext/selector.h>
#include <gtkmmext/gtk_ui.h>

#include "editor.h"
#include "config.h"
#include "tape.h"
#include "track.h"
#include "utils.h"

using namespace ARDOUR;
using namespace SigC;

Editor::TrackView::TrackView (Editor *ed, size_t tn, bool with_side_bar)
	: editor (ed),
	  tracknum (tn),
	  mute_button ("mute"),
	  solo_button ("solo"),
	  history_button ("edl"),
	  history_selector (_refill_history_list, this, 1, true, true)
	
{
	char history_title[32];

	fakeHistoryDisplay.selector = &history_selector;

	history_selector.clist().set_name ("HistorySelector");
	history_selector.selection_made.connect (slot (this, &TrackView::random_history_selection));
	
	sprintf (history_title, "ardour: track %u history", tracknum + 1);

	history_display.add (history_selector);
	history_display.set_title (history_title);
	history_display.set_usize (450, 150);
	history_display.delete_event.connect (slot (this, &TrackView::history_hide));

	waveview = gtk_wave_view_new (NULL, FALSE, FALSE);
	waveview_buffer = 0;

	gtk_widget_set_name (waveview, "Viewer");
	gtk_widget_set_name (GTK_WAVE_VIEW(waveview)->area, "Viewer");

	/* wrap the things we want to be able to use easily from Gtk-- */

	wave = Gtk::wrap (GTK_WIDGET(waveview));
	wavearea = Gtk::wrap (GTK_DRAWING_AREA(GTK_WAVE_VIEW(waveview)->area));
	x_adjuster = Gtk::wrap (gtk_wave_view_get_x_adjustment (GTK_WAVE_VIEW(waveview)));

	wavearea->realize.connect (bind (slot (Editor::set_cross_hair_cursor), wavearea));

	gtk_signal_connect (GTK_OBJECT(waveview), "cursor_set", 
			    GTK_SIGNAL_FUNC (Editor::_cursor_set_handler),
			    editor);
	
	if (with_side_bar) {
		mute_button.set_name ("TrackMuteButton");
		solo_button.set_name ("TrackSoloButton");
		edit_button.set_name ("TrackEditButton");
		history_button.set_name ("TrackHistoryButton");
		
		mute_button.toggled.connect (bind (slot (editor, &Editor::mute_toggle), tracknum));
		solo_button.toggled.connect (bind (slot (editor, &Editor::solo_toggle), tracknum));
		edit_button.button_release_event.connect (bind (slot (editor, &Editor::edit_click), tracknum));
		history_button.clicked.connect (slot (this, &TrackView::history_click));
		
		edit_label.set_text ("group");
		edit_label.set_name ("TrackEditIndicator0");
		Gtkmmext::set_usize_to_display_given_text (edit_label, "group B", 2, 2);
		
		edit_button.add (edit_label);
		edit_button.set_name ("TrackEditIndicator0");
		
		button_hbox.pack_start (solo_button, false, false);
		button_hbox.pack_start (mute_button, false, false);
		button_hbox.pack_start (edit_button, false, false);
		button_hbox.pack_end (history_button, false, false);

		track_name_display.set_name ("EditorTrackNameDisplay");
		track_name_hbox.pack_start (track_name_display, true, true);
		track_name_frame.add (track_name_hbox);
		track_name_frame.set_shadow_type (GTK_SHADOW_IN);
		track_name_frame.set_name ("BaseFrame");
		
		take_name_display.set_name ("EditorTakeNameDisplay");
		take_name_hbox.pack_start (take_name_display, true, true);
		take_name_frame.add (take_name_hbox);
		take_name_frame.set_shadow_type (GTK_SHADOW_IN);
		take_name_frame.set_name ("BaseFrame");
		
		left_vbox.pack_start (track_name_frame, false, false); 
		left_vbox.pack_start (take_name_frame, false, false); 
		left_vbox.pack_start (button_hbox, false, false, 5);

		/* remove focus from the buttons */

		solo_button.unset_flags (GTK_CAN_FOCUS);
		mute_button.unset_flags (GTK_CAN_FOCUS);
		edit_button.unset_flags (GTK_CAN_FOCUS);
		history_button.unset_flags (GTK_CAN_FOCUS);
	}


	wave_frame.set_shadow_type (GTK_SHADOW_IN);
	wave_frame.set_name ("BaseFrame");
	gtk_container_add (GTK_CONTAINER(wave_frame.gtkobj()), waveview);

	hpacker.pack_start (left_vbox, false, false, 5);
	hpacker.pack_start (wave_frame, true, true, 0);

	/* package the separator so that it can be used if desired */

	separator.set_name ("TrackSeparator");
	sep_box.pack_start (separator, true, true);

	pack_start (hpacker, true, true, 2);
}

void
Editor::TrackView::_wave_modified (GtkObject *object, GRange *range, gpointer data)

{
	((TrackView *) data)->wave_modified (range);
}

void
Editor::TrackView::wave_modified (GRange *range)

{
	track->invalidate_buffers ();
	Gtkmmext::UI::instance()->touch_display (&fakeHistoryDisplay);
}

void
Editor::TrackView::set_buffer (GtkObject *obj)

{
	waveview_buffer = obj;
	gtk_wave_view_set_buffer (GTK_WAVE_VIEW(waveview), GTK_WAVE_BUFFER(obj));
	gtk_signal_connect (obj, "modified", GTK_SIGNAL_FUNC(_wave_modified), this);
}

void
Editor::TrackView::reset (Tape *tape)

{
	GtkObject *buffer;

	if (waveview_buffer) {
		// find C "destructor" for wave buffer
	}

	track = tape->track (tracknum);
	
	buffer = gtk_edl_wave_buffer_new (track->sequence(), tape->frame_rate());

	set_buffer (buffer);
	
	/* A zoom of 256 seems to be about right */

	gtk_wave_view_set_zoom (GTK_WAVE_VIEW(waveview), 256);

	edit_group = tracknum;
	mute_button.set_active (false);
	solo_button.set_active (false);
	track_name_display.set_text (tape->track (tracknum)->tagname());
	// take_name_display.set_text (tape->track (tracknum)->take_name());
}

void
Editor::TrackView::history_click ()

{
	if (history_display.is_visible()) {
		history_display.hide_all ();
	} else {
		history_display.show_all ();
	}
}

gint
Editor::TrackView::history_hide (GdkEventAny *ev)

{
	history_display.hide_all ();
	return TRUE;
}


/* History Handling */

void
Editor::TrackView::_refill_history_list (Gtk::CList &list, void *arg)

{
	((Editor::TrackView *) arg)->refill_history_list (list);
}

void
Editor::TrackView::refill_history_list (Gtk::CList &list)
{
	GList *state;
	const gchar *rowdata[1];
	char buf[1024];
	int row = 0;
	sequence_t *seq;
	int current;

	/* XXX the region trackview has tracknum < 0 and no edit
	   history FIX ME
	*/

	if (editor->tape == 0 || tracknum < 0) {
		return;
	}

	seq = editor->tape->track (tracknum)->sequence();
	current = sequence_state_stack_position (seq);

	list.set_column_auto_resize (0, true);
	list.set_column_auto_resize (1, true);

	for (state = g_list_last (seq->state_ptr); state; state = g_list_previous (state)) {
		format_edl_operation (((sequence_state_t *) state->data)->op, 
				      buf,
				      editor->tape->frame_rate(),
				      Config->smpte_frames_per_second,
				      sizeof (ARDOUR::Sample),
				      true);
		rowdata[0] = buf;
		list.insert_row (row++, rowdata);
	}

	if (row) {
		printf ("current position in history = %d, rowcnt = %d\n", current, row);
	
		list.row (row - current - 1).select ();
		list.moveto (row - current - 1, 0, 0.5, 0);
	}
}

void
Editor::TrackView::random_history_selection (Gtkmmext::Selector *selector,
					     Gtkmmext::SelectionResult *res)

{
	sequence_t *seq;

	if (editor->tape == 0 || tracknum < 0) {
		return;
	}

	seq = gtk_edl_wave_buffer_get_sequence (GTK_EDITABLE_WAVE_BUFFER(waveview_buffer));
	sequence_set_state_stack_position (seq, selector->clist().rows().size() - res->row);
}


